__copyright__ = "Copyright (C) 2019 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging

import numpy as np
import pytest

import pyopencl as cl
import pyopencl.clmath
import pyopencl.clrandom
from pymbolic.mapper.evaluator import EvaluationMapper
from pyopencl.tools import (  # noqa: F401
    pytest_generate_tests_for_pyopencl as pytest_generate_tests,
)

import loopy as lp
from loopy.version import LOOPY_USE_LANGUAGE_VERSION_2018_2  # noqa: F401


logger = logging.getLogger(__name__)


# {{{ code generator fuzzing

class BoundsCheckError(ValueError):
    pass


class BoundsCheckingEvaluationMapper(EvaluationMapper):
    def __init__(self, context, lbound, ubound):
        super().__init__(context)
        self.lbound = lbound
        self.ubound = ubound

    def rec(self, expr):
        result = super().rec(expr)

        if result > self.ubound:
            raise BoundsCheckError()
        if result < self.lbound:
            raise BoundsCheckError()

        return result


def make_random_fp_value(use_complex):
    from random import randrange, uniform
    v = randrange(3)
    if v == 0:
        while True:
            z = randrange(-1000, 1000)
            if z:
                return z

    elif v == 1 or not use_complex:
        return uniform(-10, 10)
    else:
        cval = uniform(-10, 10) + 1j*uniform(-10, 10)
        if randrange(0, 2) == 0:
            return np.complex128(cval)
        else:
            return np.complex128(cval)


def make_random_fp_expression(prefix, var_values, size, use_complex):
    from random import randrange

    import pymbolic.primitives as p
    v = randrange(1500)
    size[0] += 1
    if v < 500 and size[0] < 40:
        term_count = randrange(2, 5)
        if randrange(2) < 1:
            cls = p.Sum
        else:
            cls = p.Product
        return cls(tuple(
            make_random_fp_expression(prefix, var_values, size, use_complex)
            for i in range(term_count)))
    elif v < 750:
        return make_random_fp_value(use_complex=use_complex)
    elif v < 1000:
        var_name = "var_%s_%d" % (prefix, len(var_values))
        assert var_name not in var_values
        var_values[var_name] = make_random_fp_value(use_complex=use_complex)
        return p.Variable(var_name)
    elif v < 1250:
        # FIXME: What does this comment mean?
        # Cannot use '-' because that destroys numpy constants.
        return p.Sum((
            make_random_fp_expression(prefix, var_values, size, use_complex),
            - make_random_fp_expression(prefix, var_values, size, use_complex)))
    elif v < 1500:
        # FIXME: What does this comment mean?
        # Cannot use '/' because that destroys numpy constants.
        return p.Quotient(
                make_random_fp_expression(prefix, var_values, size, use_complex),
                make_random_fp_expression(prefix, var_values, size, use_complex))
    else:
        raise AssertionError()


def make_random_int_value(nonneg):
    from random import randrange
    if nonneg:
        return randrange(1, 100)

    else:
        while True:
            z = randrange(-100, 100)
            if z:
                return z


def make_random_int_expression(prefix, var_values, size, nonneg):
    from random import randrange

    import pymbolic.primitives as p
    if size[0] < 10:
        v = randrange(800)
    else:
        v = randrange(1000)

    size[0] += 1

    if v < 10:
        var_name = "var_%s_%d" % (prefix, len(var_values))
        assert var_name not in var_values
        var_values[var_name] = make_random_int_value(nonneg)
        return p.Variable(var_name)
    elif v < 200 and size[0] < 40:
        term_count = randrange(2, 5)
        if randrange(2) < 1:
            cls = p.Sum
        else:
            cls = p.Product
        return cls(tuple(
            make_random_int_expression(
                prefix, var_values, size, nonneg=nonneg)
            for i in range(term_count)))
    elif v < 400 and size[0] < 40:
        return p.FloorDiv(
                make_random_int_expression(
                    prefix, var_values, size, nonneg=nonneg),
                make_nonzero_random_int_expression(
                    prefix, var_values, size, nonneg=nonneg))
    elif v < 600 and size[0] < 40:
        return p.Remainder(
                make_random_int_expression(
                    prefix, var_values, size, nonneg=nonneg),
                make_nonzero_random_int_expression(
                    prefix, var_values, size, nonneg=nonneg))
    elif v < 800 and not nonneg and size[0] < 40:
        return p.Sum((
            make_random_int_expression(
                prefix, var_values, size, nonneg=nonneg),
            - make_random_int_expression(
                prefix, var_values, size, nonneg=nonneg),
            ))
    else:
        return make_random_int_value(nonneg)


def make_nonzero_random_int_expression(prefix, var_values, size, nonneg):
    while True:
        var_values_new = var_values.copy()
        size_new = size[:]
        result = make_random_int_expression(
                prefix, var_values_new, size_new, nonneg)

        result_eval = EvaluationMapper(var_values_new)(result)

        if result_eval != 0:
            var_values.update(var_values_new)
            size[:] = size_new

            return result

    raise AssertionError()


def generate_random_fuzz_examples(expr_type):
    i = 0
    while True:
        size = [0]
        var_values = {}
        if expr_type in ["real", "complex"]:
            expr = make_random_fp_expression("e%d" % i, var_values, size,
                    use_complex=expr_type == "complex")
        elif expr_type == "int":
            expr = make_random_int_expression(
                    "e%d" % i, var_values, size, nonneg=False)
        elif expr_type == "int_nonneg":
            expr = make_random_int_expression(
                    "e%d" % i, var_values, size, nonneg=True)
        else:
            raise ValueError("unknown expr_type: %s" % expr_type)

        yield i, expr, var_values

        i += 1


def assert_parse_roundtrip(expr):
    from pymbolic.mapper.stringifier import StringifyMapper
    strified = StringifyMapper()(expr)
    from pymbolic import parse
    parsed_expr = parse(strified)
    print(expr)
    print(parsed_expr)
    assert expr == parsed_expr


@pytest.mark.parametrize("target_cls", [lp.PyOpenCLTarget, lp.ExecutableCTarget])
@pytest.mark.parametrize("random_seed", [0, 1, 2, 3, 4, 5])
@pytest.mark.parametrize("expr_type", ["int", "int_nonneg", "real", "complex"])
def test_fuzz_expression_code_gen(ctx_factory, expr_type, random_seed, target_cls):
    from pymbolic import evaluate

    def get_numpy_type(x):
        if expr_type in ["real", "complex"]:
            if isinstance(x, (complex, np.complexfloating)):
                return np.complex128
            else:
                return np.float64

        elif expr_type in ["int", "int_nonneg"]:
            return np.int64

        else:
            raise ValueError("unknown expr_type: %s" % expr_type)

    from random import seed

    seed(random_seed)

    data = []
    instructions = []

    ref_values = {}

    if expr_type in ["real", "complex"]:
        result_type = np.complex128
    elif expr_type in ["int", "int_nonneg"]:
        result_type = np.int64
    else:
        raise AssertionError()

    var_names = []

    fuzz_iter = iter(generate_random_fuzz_examples(expr_type))
    count = 0

    while True:
        if count == 10:
            break

        i, expr, var_values = next(fuzz_iter)

        var_name = "expr%d" % i

        # print(expr)
        # assert_parse_roundtrip(expr)

        if expr_type in ["int", "int_nonneg"]:
            result_type_iinfo = np.iinfo(np.int32)
            bceval_mapper = BoundsCheckingEvaluationMapper(
                    var_values,
                    lbound=result_type_iinfo.min,
                    ubound=result_type_iinfo.max)
            # print(expr)
            try:
                ref_values[var_name] = bceval_mapper(expr)
            except BoundsCheckError:
                print(expr)
                print("BOUNDS CHECK FAILED")
                continue
        else:
            try:
                ref_values[var_name] = evaluate(expr, var_values)
            except ZeroDivisionError:
                continue

        count += 1

        data.append(lp.GlobalArg(var_name,
            result_type,
            shape=()))
        data.extend([
            lp.TemporaryVariable(name, get_numpy_type(val))
            for name, val in var_values.items()
            ])
        instructions.extend([
            lp.Assignment(name, get_numpy_type(val)(val))
            for name, val in var_values.items()
            ])
        instructions.append(lp.Assignment(var_name, expr))

        if expr_type == "int_nonneg":
            var_names.extend(var_values)

    if issubclass(target_cls, lp.ExecutableCTarget):
        # https://github.com/inducer/loopy/issues/686
        # https://gcc.gnu.org/bugzilla/show_bug.cgi?id=107127

        from shutil import which
        gcc_10 = which("gcc-10")
        if gcc_10 is not None:
            from loopy.target.c.c_execution import CCompiler
            target = target_cls(compiler=CCompiler(cc=gcc_10))
        else:
            from warnings import warn
            warn("Using default C compiler because gcc-10 was not found. "
                 "These tests may take a long time, because of "
                 "https://gcc.gnu.org/bugzilla/show_bug.cgi?id=107127.",
                 stacklevel=1)
            target = target_cls()

    else:
        target = target_cls()

    knl = lp.make_kernel("{ : }", instructions, data, seq_dependencies=True,
            target=target)

    import islpy as isl
    knl = lp.assume(knl, isl.BasicSet(
            "[%s] -> { : %s}"
            % (
                ", ".join(var_names),
                " and ".join("%s >= 0" % name for name in var_names))))

    knl = lp.set_options(knl, return_dict=True)
    # print(knl)

    if type(target) is lp.PyOpenCLTarget:
        cl_ctx = ctx_factory()
        knl = lp.set_options(knl, write_code=True)
        with cl.CommandQueue(cl_ctx) as queue:
            _evt, lp_values = knl(queue, out_host=True)
    elif type(target) is lp.ExecutableCTarget:
        _evt, lp_values = knl()
    else:
        raise NotImplementedError("unsupported target")

    for name, ref_value in ref_values.items():
        lp_value = lp_values[name]
        if expr_type in ["real", "complex"]:
            err = abs(ref_value-lp_value)/abs(ref_value)
        elif expr_type in ["int", "int_nonneg"]:
            err = abs(ref_value-lp_value)
        else:
            raise AssertionError()

        if abs(err) > 1e-10:
            print(80*"-")
            print(knl)
            print(80*"-")
            print(lp.generate_code_v2(knl).device_code())
            print(80*"-")
            print(f"WRONG: {name} rel error={err:g}")
            print("reference=%r" % ref_value)
            print("loopy=%r" % lp_value)
            print(80*"-")
            1/0  # noqa: B018

    print(lp.generate_code_v2(knl).device_code())

# }}}


def test_sci_notation_literal(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    set_kernel = lp.make_kernel(
         """ { [i]: 0<=i<12 } """,
         """ out[i] = 1e-12""")

    set_kernel = lp.set_options(set_kernel, write_code=True)

    _evt, (out,) = set_kernel(queue)

    assert (np.abs(out.get() - 1e-12) < 1e-20).all()


def test_indexof(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    knl = lp.make_kernel(
         """ { [i,j]: 0<=i,j<5 } """,
         """ out[i,j] = indexof(out[i,j])""",
         [lp.GlobalArg("out", is_input=False, shape=lp.auto)]
    )

    knl = lp.set_options(knl, write_code=True)

    (_evt, (out,)) = knl(queue)
    out = out.get()

    assert np.array_equal(out.ravel(order="C"), np.arange(25))


def test_indexof_vec(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    if (
            # Accurate as of 2019-11-04
            ctx.devices[0].platform.name.startswith("Intel")):
        pytest.skip("target ICD miscompiles vector code")

    knl = lp.make_kernel(
         """ { [i,j,k]: 0<=i,j,k<4 } """,
         """ out[i,j,k] = indexof_vec(out[i,j,k])""",
         [lp.GlobalArg("out", shape=lp.auto, is_input=False)])

    knl = lp.tag_inames(knl, {"i": "vec"})
    knl = lp.tag_data_axes(knl, "out", "vec,c,c")
    knl = lp.set_options(knl, write_code=True)

    (_evt, (_out,)) = knl(queue)
    # out = out.get()
    # assert np.array_equal(out.ravel(order="C"), np.arange(25))


def test_is_expression_equal():
    from pymbolic import var

    from loopy.symbolic import is_expression_equal

    x = var("x")
    y = var("y")

    assert is_expression_equal(x+2, 2+x)

    assert is_expression_equal((x+2)**2, x**2 + 4*x + 4)
    assert is_expression_equal((x+y)**2, x**2 + 2*x*y + y**2)


def test_integer_associativity():
    knl = lp.make_kernel(
            "{[i] : 0<=i<arraylen}",
            """
            e := (i // (ncomp * elemsize))
            d := ((i // elemsize) % ncomp)
            s := (i % elemsize)
            v[i] = u[ncomp * indices[(s) + elemsize*(e)] + (d)]
            """)

    knl = lp.add_and_infer_dtypes(
            knl, {"u": np.float64, "elemsize, ncomp, indices": np.int32})
    import islpy as isl
    knl = lp.assume(
            knl, isl.BasicSet("[elemsize, ncomp] -> "
            "{ : elemsize>= 0 and ncomp >= 0}"))
    print(lp.generate_code_v2(knl).device_code())
    assert (
            "u[ncomp * indices[i % elemsize + elemsize "
            "* (i / (ncomp * elemsize))] "
            "+ loopy_mod_pos_b_int32(i / elemsize, ncomp)]"
            in lp.generate_code_v2(knl).device_code())


def test_floor_div():
    knl = lp.make_kernel(
        "{ [i]: 0<=i<10 }",
        "out[i] = (i-1)*(i-2)//2")
    assert "loopy_floor_div" in lp.generate_code_v2(knl).device_code()

    knl = lp.make_kernel(
        "{ [i]: 0<=i<10 }",
        "out[i] = (i*(i+1))//2")
    assert "loopy_floor_div" not in lp.generate_code_v2(knl).device_code()


def test_divide_precedence(ctx_factory):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    knl = lp.make_kernel(
            "{:}",
            """
            x[0] = c*(a/b)
            y[0] = c*(a%b)
            """,
            [lp.ValueArg("a, b, c", np.int32),
                lp.GlobalArg("x, y", np.int32, shape=lp.auto)])
    print(lp.generate_code_v2(knl).device_code())

    evt, (x_out, y_out) = knl(queue, c=2, b=2, a=5)
    evt.wait()
    assert x_out.get() == 4
    assert y_out.get() == 2


@pytest.mark.parametrize("target", [lp.PyOpenCLTarget, lp.ExecutableCTarget])
def test_complex_support(ctx_factory, target):
    knl = lp.make_kernel(
            "{[i, i1, i2]: 0<=i,i1,i2<10}",
            """
            euler1[i]  = exp(3.14159265359j)
            euler2[i] = (2.7182818284 ** (3.14159265359j))
            euler1_real[i] = real(euler1[i])
            euler1_imag[i] = imag(euler1[i])
            real_times_complex[i] = in1[i]*(in2[i]*1j)
            real_plus_complex[i] = in1[i] + (in2[i]*1j)
            abs_complex[i] = abs(real_plus_complex[i])
            complex_div_complex[i] = (2jf + 7*in1[i])/(32jf + 37*in1[i])
            complex_div_real[i] = (2jf + 7*in1[i])/in1[i]
            real_div_complex[i] = in1[i]/(2jf + 7*in1[i])
            out_sum = sum(i1, 1.0*i1 + i1*1jf)*sum(i2, 1.0*i2 + i2*1jf)
            conj_out_sum = conj(out_sum)
            """,
            [
                lp.GlobalArg("out_sum, euler1, real_plus_complex",
                            is_input=False, shape=lp.auto),
                ...
            ],
            target=target(),
            seq_dependencies=True)
    knl = lp.set_options(knl, return_dict=True)

    n = 10
    rng = np.random.default_rng(seed=42)

    in1 = rng.random(n)
    in2 = rng.random(n)

    kwargs = {"in1": in1, "in2": in2}

    knl = lp.set_options(knl, write_code=True)

    if target == lp.PyOpenCLTarget:
        cl_ctx = ctx_factory()
        with cl.CommandQueue(cl_ctx) as queue:
            _evt, out = knl(queue, **kwargs)
    elif target == lp.ExecutableCTarget:
        _evt, out = knl(**kwargs)
    else:
        raise NotImplementedError("unsupported target")

    np.testing.assert_allclose(out["euler1"], -1)
    np.testing.assert_allclose(out["euler2"], -1)
    np.testing.assert_allclose(out["euler1_real"], -1)
    np.testing.assert_allclose(out["euler1_imag"], 0, atol=1e-10)
    np.testing.assert_allclose(out["real_times_complex"], in1*(in2*1j))
    np.testing.assert_allclose(out["real_plus_complex"], in1+(in2*1j))
    np.testing.assert_allclose(out["abs_complex"], np.abs(out["real_plus_complex"]))
    np.testing.assert_allclose(out["complex_div_complex"], (2j+7*in1)/(32j+37*in1))
    np.testing.assert_allclose(out["complex_div_real"], (2j + 7*in1)/in1)
    np.testing.assert_allclose(out["real_div_complex"], in1/(2j + 7*in1))
    np.testing.assert_allclose(out["out_sum"], (0.5*n*(n-1) + 0.5*n*(n-1)*1j) ** 2)
    np.testing.assert_allclose(out["conj_out_sum"],
                               (0.5*n*(n-1) - 0.5*n*(n-1)*1j) ** 2)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_real_with_real_argument(ctx_factory, dtype):
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    knl = lp.make_kernel(
            "{[i]: 0 <= i < nresult}",
            "result[i] = real(ary[i])",
            )

    rng = np.random.default_rng()
    ary = cl.array.to_device(queue, rng.random(128).astype(dtype))

    _, (result,) = knl(queue, ary=ary)

    assert result.dtype == ary.dtype
    np.testing.assert_allclose(result.get(), np.real(ary.get()))


def test_bool_type_context(ctx_factory):
    # Checks if a boolean type context is correctly handled in codegen phase.
    # See https://github.com/inducer/loopy/pull/258
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    knl = lp.make_kernel(
        "{:}",
        """
        k = 8.0 and 7.0
        """,
        [
            lp.GlobalArg("k", dtype=np.bool_, shape=lp.auto),
        ])

    _evt, (out,) = knl(queue)
    assert out.get() == np.logical_and(7.0, 8.0)


def test_np_bool_handling(ctx_factory):
    import pymbolic.primitives as p

    from loopy.symbolic import parse
    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    knl = lp.make_kernel(
        "{:}",
        [lp.Assignment(parse("y"), p.LogicalNot(np.bool_(False)))],
        [lp.GlobalArg("y", dtype=np.bool_, shape=lp.auto)])
    _evt, (out,) = knl(queue)
    assert out.get().item() is True


@pytest.mark.parametrize("target", [lp.PyOpenCLTarget, lp.ExecutableCTarget])
def test_complex_functions_with_real_args(ctx_factory, target):
    # Reported by David Ham. See <https://github.com/inducer/loopy/issues/851>
    t_unit = lp.make_kernel(
            "{[i]: 0<=i<10}",
            """
            y1[i] = abs(c64[i])
            y2[i] = real(c64[i])
            y3[i] = imag(c64[i])
            y4[i] = conj(c64[i])

            y5[i] = abs(c128[i])
            y6[i] = real(c128[i])
            y7[i] = imag(c128[i])
            y8[i] = conj(c128[i])


            y9[i] = abs(f32[i])
            y10[i] = real(f32[i])
            y11[i] = imag(f32[i])
            y12[i] = conj(f32[i])

            y13[i] = abs(f64[i])
            y14[i] = real(f64[i])
            y15[i] = imag(f64[i])
            y16[i] = conj(f64[i])
            """,
            target=target())

    t_unit = lp.add_dtypes(t_unit,
                           {"y9,y10,y11,y12": np.complex64,
                            "y13,y14,y15,y16": np.complex128,
                            "c64": np.complex64,
                            "c128": np.complex128,
                            "f64": np.float64,
                            "f32": np.float32})
    t_unit = lp.set_options(t_unit, return_dict=True)

    from numpy.random import default_rng
    rng = default_rng(0)
    c64 = (rng.random(10, dtype=np.float32)
           + np.csingle(1j)*rng.random(10, dtype=np.float32))
    c128 = (rng.random(10, dtype=np.float64)
            + np.cdouble(1j)*rng.random(10, dtype=np.float64))
    f32 = rng.random(10, dtype=np.float32)
    f64 = rng.random(10, dtype=np.float64)

    if target == lp.PyOpenCLTarget:
        cl_ctx = ctx_factory()
        with cl.CommandQueue(cl_ctx) as queue:
            _evt, out = t_unit(queue, c64=c64, c128=c128, f32=f32, f64=f64)
    elif target == lp.ExecutableCTarget:
        t_unit = lp.set_options(t_unit, build_options=["-Werror"])
        _evt, out = t_unit(c64=c64, c128=c128, f32=f32, f64=f64)
    else:
        raise NotImplementedError("unsupported target")

    np.testing.assert_allclose(out["y1"], np.abs(c64), rtol=1e-6)
    np.testing.assert_allclose(out["y2"], np.real(c64), rtol=1e-6)
    np.testing.assert_allclose(out["y3"], np.imag(c64), rtol=1e-6)
    np.testing.assert_allclose(out["y4"], np.conj(c64), rtol=1e-6)
    np.testing.assert_allclose(out["y5"], np.abs(c128), rtol=1e-6)
    np.testing.assert_allclose(out["y6"], np.real(c128), rtol=1e-6)
    np.testing.assert_allclose(out["y7"], np.imag(c128), rtol=1e-6)
    np.testing.assert_allclose(out["y8"], np.conj(c128), rtol=1e-6)
    np.testing.assert_allclose(out["y9"], np.abs(f32), rtol=1e-6)
    np.testing.assert_allclose(out["y10"], np.real(f32), rtol=1e-6)
    np.testing.assert_allclose(out["y11"], np.imag(f32), rtol=1e-6)
    np.testing.assert_allclose(out["y12"], np.conj(f32), rtol=1e-6)
    np.testing.assert_allclose(out["y13"], np.abs(f64), rtol=1e-6)
    np.testing.assert_allclose(out["y14"], np.real(f64), rtol=1e-6)
    np.testing.assert_allclose(out["y15"], np.imag(f64), rtol=1e-6)
    np.testing.assert_allclose(out["y16"], np.conj(f64), rtol=1e-6)


if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from pytest import main
        main([__file__])

# vim: foldmethod=marker



import numpy as np


# Inspired by a visualization used in the Halide tutorial
# https://www.youtube.com/watch?v=3uiEyEKji0M


def div_ceil(nr, dr):
    return -(-nr // dr)


def product(iterable):
    from functools import reduce
    from operator import mul
    return reduce(mul, iterable, 1)


class ArrayAccessPatternContext:
    def __init__(self, gsize, lsize, subgroup_size=32, decay_constant=0.75):
        self.lsize = lsize
        self.gsize = gsize
        self.subgroup_size = subgroup_size
        self.timestamp = 0
        self.decay_constant = decay_constant

        self.ind_length = len(gsize) + len(lsize)

        self.arrays = []

    def l(self, index):  # noqa: E743
        subscript = [np.newaxis] * self.ind_length
        subscript[len(self.gsize) + index] = slice(None)

        return np.arange(self.lsize[index])[tuple(subscript)]

    def g(self, index):
        subscript = [np.newaxis] * self.ind_length
        subscript[index] = slice(None)

        return np.arange(self.gsize[index])[tuple(subscript)]

    def nsubgroups(self):
        return div_ceil(product(self.lsize), self.subgroup_size)

    def animate(self, f, interval=200):
        import matplotlib.animation as animation
        import matplotlib.pyplot as plt

        fig = plt.figure()

        plots = []
        for iary, ary in enumerate(self.arrays):
            ax = fig.add_subplot(1, len(self.arrays), 1+iary)
            ax.set_title(ary.name)
            plots.append(ary.plot(ax))

        def data_gen():
            for _ in f():
                self.tick()

                for ary, plot in zip(self.arrays, plots):
                    plot.set_array(ary.get_plot_data())

                fig.canvas.draw()
                yield plots

        # must be kept alive until after plt.show()
        return animation.FuncAnimation(
                fig, lambda x: x, data_gen,
                blit=False, interval=interval, repeat=True)

    def tick(self):
        self.timestamp += 1


class Array:
    def __init__(self, ctx, name, shape, strides, elements_per_row=None):
        # Each array element stores a tuple:
        # (timestamp, subgroup, g0, g1, g2, ) of last access

        assert len(shape) == len(strides)

        self.nattributes = 2+len(ctx.gsize)

        if elements_per_row is None:
            if len(shape) > 1:
                minstride = min(strides)
                for sh_i, st_i in zip(shape, strides):
                    if st_i == minstride:
                        elements_per_row = sh_i
                        break
        else:
            elements_per_row = 256

        self.array = np.zeros((product(shape), self.nattributes,), dtype=np.int32)

        self.ctx = ctx
        self.name = name
        self.shape = shape
        self.strides = strides
        self.elements_per_row = elements_per_row

        ctx.arrays.append(self)

    def __getitem__(self, index):
        if not isinstance(index, tuple):
            index = (index,)

        assert len(index) == len(self.shape)

        all_subscript = (np.newaxis,) * self.ctx.ind_length

        def reshape_ind(ind):
            if not isinstance(ind, np.ndarray):
                return ind[all_subscript]

            else:
                assert len(ind.shape) == self.ctx.ind_length

        lin_index = sum(
                ind_i * stride_i
                for ind_i, stride_i in zip(index, self.strides))

        if not isinstance(lin_index, np.ndarray):
            subscript = [np.newaxis] * self.ctx.ind_length
            lin_index = np.array(lin_index)[subscript]

        self.array[lin_index, 0] = self.ctx.timestamp
        for i, _glength in enumerate(self.ctx.gsize):
            if lin_index.shape[i] > 1:
                self.array[lin_index, 2+i] = self.ctx.g(i)

        workitem_index = 0
        for i in range(len(self.ctx.lsize))[::-1]:
            workitem_index = (
                    workitem_index * self.ctx.lsize[i]
                    + self.ctx.l(i))
        subgroup = workitem_index//self.ctx.subgroup_size
        self.array[lin_index, 1] = subgroup

    def __setitem__(self, index, value):
        self.__getitem__(index)

    def get_plot_data(self):
        nelements = self.array.shape[0]
        base_shape = (
                div_ceil(nelements, self.elements_per_row),
                self.elements_per_row,)
        shaped_array = np.zeros(
                (*base_shape, self.nattributes),
                dtype=np.float32)
        shaped_array.reshape(-1, self.nattributes)[:nelements] = self.array

        modulation = np.exp(
                -self.ctx.decay_constant*(self.ctx.timestamp-shaped_array[:, :, 0]))

        subgroup = shaped_array[:, :, 1]
        if self.ctx.nsubgroups() > 1:
            subgroup = subgroup/(self.ctx.nsubgroups()-1)
        else:
            subgroup.fill(1)

        rgb_array = np.zeros((*base_shape, 3))
        if 1:
            if len(self.ctx.gsize) > 1:
                # g.0 -> red
                rgb_array[:, :, 0] = shaped_array[:, :, 2]/(self.ctx.gsize[0]-1)
            if len(self.ctx.gsize) > 1:
                # g.1 -> blue
                rgb_array[:, :, 2] = shaped_array[:, :, 3]/(self.ctx.gsize[1]-1)
        if 1:
            rgb_array[:, :, 1] = subgroup

        return rgb_array*modulation[:, :, np.newaxis]

    def plot(self, ax, **kwargs):
        return ax.imshow(
                self.get_plot_data(), interpolation="nearest",
                **kwargs)


def show_example():
    n = 2**7
    n16 = div_ceil(n, 16)
    ctx = ArrayAccessPatternContext(gsize=(n16, n16), lsize=(16, 16))
    in0 = Array(ctx, "in0", (n, n), (n, 1))

    if 0:
        # knl a
        i_inner = ctx.l(1)
        i_outer = ctx.g(1)
        k_inner = ctx.l(0)

        def f():
            for k_outer in range(n16):
                in0[i_inner + i_outer*16, k_inner + k_outer*16]
                yield
    elif 0:
        # knl b
        j_inner = ctx.l(0)
        j_outer = ctx.g(0)
        k_inner = ctx.l(1)

        def f():
            for k_outer in range(n16):
                in0[k_inner + k_outer*16, j_inner + j_outer*16]
                yield

    ani = ctx.animate(f)
    import matplotlib.pyplot as plt
    if 1:
        plt.show()
    else:
        ani.save("access.mp4")


def show_example_2():
    bsize = 8
    blocks = 3

    ctx = ArrayAccessPatternContext(gsize=(1,), lsize=(1,),
            decay_constant=0.005)
    in0 = Array(ctx, "in0", (blocks*bsize, blocks*bsize), (blocks*bsize, 1))

    def f():
        for i_outer in range(blocks):
            for j_outer in range(blocks):
                for i_inner in range(bsize):
                    for j_inner in range(bsize):
                        in0[i_inner + i_outer*bsize, j_inner + j_outer*bsize]
                        yield

    ani = ctx.animate(f, interval=10)
    import matplotlib.pyplot as plt
    if 1:
        plt.show()
    else:
        ani.save("access.mp4")


if __name__ == "__main__":
    show_example_2()



from __future__ import annotations


__copyright__ = "Copyright (C) 2018 Kaushik Kulkarni"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from typing import TYPE_CHECKING

from constantdict import constantdict

import islpy as isl
from pytools import UniqueNameGenerator

from loopy.diagnostic import LoopyError
from loopy.kernel import LoopKernel
from loopy.kernel.function_interface import (
    CallableKernel,
    InKernelCallable,
    ScalarCallable,
)
from loopy.kernel.instruction import (
    Assignment,
    CallInstruction,
    CInstruction,
    MultiAssignmentBase,
    _DataObliviousInstruction,
)
from loopy.symbolic import (
    RuleAwareIdentityMapper,
    RuleAwareSubstitutionMapper,
    SubstitutionRuleMappingContext,
)
from loopy.translation_unit import FunctionIdT, TranslationUnit, for_each_kernel


if TYPE_CHECKING:
    from collections.abc import Sequence


__doc__ = """
.. currentmodule:: loopy

.. autofunction:: register_callable

.. autofunction:: inline_callable_kernel

.. autofunction:: merge
"""


def register_callable(translation_unit, function_identifier, callable_,
        redefining_not_ok=True):
    """
    :param translation_unit: A :class:`loopy.TranslationUnit`.
    :param callable_: A :class:`loopy.InKernelCallable`.
    """

    if isinstance(callable_, LoopKernel):
        callable_ = CallableKernel(callable_)

    from loopy.kernel.function_interface import InKernelCallable
    assert isinstance(callable_, InKernelCallable)

    if (function_identifier in translation_unit.callables_table) and (
            translation_unit.callables_table[function_identifier] != callable_
            and redefining_not_ok):
        raise LoopyError("Redefining function identifier not allowed. Set the"
                " option 'redefining_not_ok=False' to bypass this error.")

    new_callables = translation_unit.callables_table.set(function_identifier,
            callable_)

    return translation_unit.copy(
            callables_table=new_callables)


def merge(translation_units: Sequence[TranslationUnit]) -> TranslationUnit:
    """
    :returns: An instance of :class:`loopy.TranslationUnit` which contains all the
        callables from each of the *translation_units.
    """

    for i in range(1, len(translation_units)):
        if translation_units[i].target != translation_units[i-1].target:
            raise LoopyError("translation units to be merged should have the"
                             " same target.")

    # {{{ check for callable collision

    for i, prg_i in enumerate(translation_units):
        for prg_j in translation_units[i+1:]:
            for clbl_name in (set(prg_i.callables_table)
                              & set(prg_j.callables_table)):
                if (prg_i.callables_table[clbl_name]
                        != prg_j.callables_table[clbl_name]):
                    # TODO: generate unique names + rename for the colliding
                    # callables (if entrypoints are colliding that should still
                    # be an error)
                    raise NotImplementedError("Translation units to be merged"
                                              " must have different callable names"
                                              " for now.")

    # }}}

    callables_table: dict[FunctionIdT, InKernelCallable] = {}
    for trans_unit in translation_units:
        callables_table.update(trans_unit.callables_table)

    return TranslationUnit(
            entrypoints=frozenset().union(*(
                t.entrypoints or frozenset() for t in translation_units)),
            callables_table=constantdict(callables_table),
            target=translation_units[0].target)


# {{{ kernel inliner mapper

class KernelArgumentSubstitutor(RuleAwareIdentityMapper):
    def __init__(self, rule_mapping_context, caller_knl,
                 callee_knl, callee_arg_to_call_param):
        super().__init__(rule_mapping_context)
        self.caller_knl = caller_knl

        # CAUTION: This kernel has post-substitution domains!
        self.callee_knl = callee_knl

        self.callee_arg_to_call_param = callee_arg_to_call_param

    def map_subscript(self, expr, expn_state):
        if expr.aggregate.name in self.callee_knl.arg_dict:
            from pymbolic import substitute
            from pymbolic.primitives import Subscript, Variable

            sar = self.callee_arg_to_call_param[expr.aggregate.name]  # SubArrayRef

            index_tuple = self.rec(expr.index_tuple, expn_state)
            subs_map = {iname: idx for idx, iname in
                    zip(index_tuple, sar.swept_inames)}
            new_indices = tuple(substitute(idx, subs_map) for idx in
                    sar.subscript.index_tuple)

            return Subscript(Variable(sar.subscript.aggregate.name), new_indices)
        else:
            return super().map_subscript(expr, expn_state)

    def map_variable(self, expr, expn_state):
        from loopy.kernel.data import ArrayArg, ValueArg
        from loopy.symbolic import SubArrayRef
        if expr.name in self.callee_knl.arg_dict:
            arg = self.callee_knl.arg_dict[expr.name]
            par = self.callee_arg_to_call_param[expr.name]
            if isinstance(arg, ArrayArg):
                assert arg.shape == ()
                assert isinstance(par, SubArrayRef) and par.swept_inames == ()
                if par.subscript.index_tuple:
                    return par.subscript
                else:
                    assert (self
                            .caller_knl
                            .get_var_descriptor(par.subscript
                                                .aggregate.name)
                            .shape) == ()
                    return par.subscript.aggregate
            else:
                assert isinstance(arg, ValueArg)
                return par
        else:
            return super().map_variable(expr, expn_state)

# }}}


# {{{ inlining of a single call instruction

def substitute_into_domain(domain, param_name, expr, allowed_param_dims):
    """
    :arg allowed_deps: A :class:`list` of :class:`str` that are
    """
    import pymbolic.primitives as prim

    from loopy.symbolic import get_dependencies, isl_set_from_expr
    if param_name not in domain.get_var_dict():
        # param_name not in domain => domain will be unchanged
        return domain

    # {{{ rename 'param_name' to avoid namespace pollution with allowed_param_dims

    dt, pos = domain.get_var_dict()[param_name]
    domain = domain.set_dim_name(dt, pos, UniqueNameGenerator(
        set(allowed_param_dims))(param_name))

    # }}}

    for dep in get_dependencies(expr):
        if dep in allowed_param_dims:
            domain = domain.add_dims(isl.dim_type.param, 1)
            domain = domain.set_dim_name(
                    isl.dim_type.param,
                    domain.dim(isl.dim_type.param)-1,
                    dep)
        else:
            raise ValueError("Augmenting caller's domain "
                    f"with '{dep}' is not allowed.")

    set_ = isl_set_from_expr(domain.space,
            prim.Comparison(prim.Variable(param_name),
                            "==",
                            expr))

    bset, = set_.get_basic_sets()
    domain = domain & bset

    return domain.project_out(dt, pos, 1)


def rename_iname(domain, old_iname, new_iname):
    if old_iname not in domain.get_var_dict():
        return domain

    dt, pos = domain.get_var_dict()[old_iname]
    return domain.set_dim_name(dt, pos, new_iname)


def get_valid_domain_param_names(knl):
    from loopy.kernel.data import ValueArg
    return ([arg.name for arg in knl.args if isinstance(arg, ValueArg)]
            + [tv.name
               for tv in knl.temporary_variables.values()
               if tv.shape == ()]
            + list(knl.all_inames())
            )


def _inline_call_instruction(caller_knl, callee_knl, call_insn):
    """
    Returns a copy of *caller_knl* with the *call_insn* in the *kernel*
    replaced by inlining *callee_knl* into it within it.

    :arg call_insn: An instance of `loopy.CallInstruction` of the call-site.
    """
    import pymbolic.primitives as prim
    from pymbolic.mapper.substitutor import make_subst_func

    from loopy.kernel.data import ValueArg

    # {{{ sanity checks

    assert call_insn.expression.function.name == callee_knl.name

    # }}}

    callee_label = callee_knl.name[:4] + "_"
    vng = caller_knl.get_var_name_generator()
    ing = caller_knl.get_instruction_id_generator()
    # collisions with callee var names might affect the renaming logic
    # below, better to avoid it.
    vng.add_names(callee_knl.all_variable_names(), conflicting_ok=True)
    ing.add_names(callee_knl.all_variable_names(), conflicting_ok=True)

    # {{{ construct callee->caller name mappings

    # name_map: Mapping[str, str]
    # A mapping from variable names in the callee kernel's namespace to
    # the ones they would be referred by in the caller's namespace post inlining.
    name_map = {}

    # only consider temporary variables and inames, arguments would be mapping
    # according to the invocation in call_insn.
    for name in (callee_knl.all_inames()
                 | set(callee_knl.temporary_variables.keys())):
        new_name = vng(callee_label+name)
        name_map[name] = new_name

    # }}}

    # {{{ iname_to_tags

    # new_inames: caller's inames post inlining
    new_inames = caller_knl.inames.copy()

    for old_name, callee_iname in callee_knl.inames.items():
        new_name = name_map[old_name]
        new_inames[new_name] = callee_iname.copy(name=new_name)

    # }}}

    # {{{ register callee's temps as caller's

    # new_temps: caller's temps post inlining
    new_temps = caller_knl.temporary_variables.copy()

    for name, tv in callee_knl.temporary_variables.items():
        new_temps[name_map[name]] = tv.copy(name=name_map[name])

    # }}}

    # {{{ get callee args -> parameters passed to the call

    arg_map = {}  # callee arg name -> caller symbols (e.g. SubArrayRef)

    assignees = call_insn.assignees  # writes
    parameters = call_insn.expression.parameters  # reads

    from loopy.kernel.function_interface import get_kw_pos_association
    _kw_to_pos, pos_to_kw = get_kw_pos_association(callee_knl)

    for i, par in enumerate(parameters):
        arg_map[pos_to_kw[i]] = par

    for i, assignee in enumerate(assignees):
        arg_map[pos_to_kw[-i-1]] = assignee

    # }}}

    # {{{ process domains/assumptions

    # rename inames
    new_domains = callee_knl.domains.copy()
    for old_iname in callee_knl.all_inames():
        new_domains = [rename_iname(dom, old_iname, name_map[old_iname])
                       for dom in new_domains]

    # realize domains' dim params in terms of caller's variables
    new_assumptions = callee_knl.assumptions
    for callee_arg_name, param_expr in arg_map.items():
        if isinstance(callee_knl.arg_dict[callee_arg_name],
                      ValueArg):
            new_domains = [
                    substitute_into_domain(
                        dom,
                        callee_arg_name,
                        param_expr, get_valid_domain_param_names(caller_knl))
                    for dom in new_domains]

            new_assumptions = substitute_into_domain(
                        new_assumptions,
                        callee_arg_name,
                        param_expr, get_valid_domain_param_names(caller_knl))

    # }}}

    # {{{ rename inames/temporaries in the program

    rule_mapping_context = SubstitutionRuleMappingContext(callee_knl.substitutions,
                                                          vng)
    subst_func = make_subst_func({old_name: prim.Variable(new_name)
                                  for old_name, new_name in name_map.items()})
    inames_temps_renamer = RuleAwareSubstitutionMapper(rule_mapping_context,
                                                       subst_func,
                                                       within=lambda *args: True)

    callee_knl = rule_mapping_context.finish_kernel(inames_temps_renamer
                                                    .map_kernel(callee_knl))

    # }}}

    # {{{ map callee's expressions to get expressions after inlining

    rule_mapping_context = SubstitutionRuleMappingContext(callee_knl.substitutions,
                                                          vng)
    smap = KernelArgumentSubstitutor(
            rule_mapping_context, caller_knl,

            # HACK: The kernel returned by this copy doesn't make sense:
            # It uses caller inames in its domain. The domains are/should be
            # only used for expression simplification. Ideally, we'd pass
            # the domains for this separately.
            # Other than that, the kernel is used for looking up argument
            # definitions, which is OK.
            callee_knl.copy(domains=new_domains),

            arg_map)

    callee_knl = rule_mapping_context.finish_kernel(smap.map_kernel(callee_knl))

    # }}}

    # {{{ generate new ids for instructions

    insn_id_map = {}
    for insn in callee_knl.instructions:
        insn_id_map[insn.id] = ing(callee_label+insn.id)

    # }}}

    # {{{ use NoOp to mark the start and end of callee kernel

    from loopy.kernel.instruction import NoOpInstruction

    noop_start = NoOpInstruction(
        id=ing(callee_label+"_start"),
        within_inames=call_insn.within_inames,
        depends_on=call_insn.depends_on,
        predicates=call_insn.predicates,
    )
    noop_end = NoOpInstruction(
        id=call_insn.id,
        within_inames=call_insn.within_inames,
        depends_on=frozenset(insn_id_map.values()),
        depends_on_is_final=True,
        predicates=call_insn.predicates,
    )

    # }}}

    # {{{ map callee's instruction ids

    inlined_insns = [noop_start]

    for insn in callee_knl.instructions:
        new_within_inames = (frozenset(name_map[iname]
                                       for iname in insn.within_inames)
                             | call_insn.within_inames)
        new_depends_on = (frozenset(insn_id_map[dep] for dep in insn.depends_on)
                          | {noop_start.id})
        new_no_sync_with = frozenset((insn_id_map[id], scope)
                                     for id, scope in insn.no_sync_with)
        new_id = insn_id_map[insn.id]

        if isinstance(insn, Assignment):
            new_atomicity = tuple(type(atomicity)(name_map[atomicity.var_name])
                                  for atomicity in insn.atomicity)
            insn = insn.copy(
                id=insn_id_map[insn.id],
                within_inames=new_within_inames,
                depends_on=new_depends_on,
                depends_on_is_final=True,
                tags=insn.tags | call_insn.tags,
                atomicity=new_atomicity,
                predicates=insn.predicates | call_insn.predicates,
                no_sync_with=new_no_sync_with
            )
        else:
            insn = insn.copy(
                id=new_id,
                within_inames=new_within_inames,
                depends_on=new_depends_on,
                tags=insn.tags | call_insn.tags,
                no_sync_with=new_no_sync_with,
                predicates=insn.predicates | call_insn.predicates,
            )
        inlined_insns.append(insn)

    inlined_insns.append(noop_end)

    # }}}

    # {{{ swap out call_insn with inlined_instructions

    idx = caller_knl.instructions.index(call_insn)
    new_insns = (caller_knl.instructions[:idx]
                 + inlined_insns
                 + caller_knl.instructions[idx+1:])

    # }}}

    old_assumptions, new_assumptions = isl.align_two(
            caller_knl.assumptions, new_assumptions)

    return caller_knl.copy(instructions=new_insns,
                           temporary_variables=new_temps,
                           domains=caller_knl.domains+new_domains,
                           assumptions=(old_assumptions.params()
                                        & new_assumptions.params()),
                           inames=new_inames,
                           preambles=caller_knl.preambles+callee_knl.preambles,
                           preamble_generators=(caller_knl.preamble_generators
                                                + callee_knl.preamble_generators),
                           )

# }}}


# {{{ inline callable kernel

@for_each_kernel
def _inline_single_callable_kernel(caller_kernel, callee_kernel):
    from loopy.symbolic import ResolvedFunction

    # sub-array refs might be removed during inlining
    # => remove their swept inames from domains
    inames_to_remove = frozenset()

    for insn in caller_kernel.instructions:
        if (isinstance(insn, CallInstruction)
                and isinstance(insn.expression.function, ResolvedFunction)):
            if insn.expression.function.name == callee_kernel.name:
                caller_kernel = _inline_call_instruction(caller_kernel,
                                                         callee_kernel, insn)
                inames_to_remove |= insn.sub_array_ref_inames()
        elif isinstance(insn, (MultiAssignmentBase, CInstruction,
                               _DataObliviousInstruction)):
            pass
        else:
            raise NotImplementedError(type(insn))

    from loopy.transform.iname import remove_unused_inames
    return remove_unused_inames(caller_kernel, inames_to_remove)


# FIXME This should take a 'within' parameter to be able to only inline
# *some* calls to a kernel, but not others.
def inline_callable_kernel(translation_unit, function_name):
    """
    Returns a copy of *translation_unit* with the callable kernel
    named *function_name* inlined at all call-sites.
    """
    from loopy.preprocess import filter_reachable_callables, infer_arg_descr
    from loopy.translation_unit import resolve_callables

    # {{{ must have argument shape information at call sites to inline

    translation_unit = resolve_callables(translation_unit)
    translation_unit = infer_arg_descr(translation_unit)

    # }}}

    callee = translation_unit[function_name]

    return filter_reachable_callables(
        _inline_single_callable_kernel(translation_unit,
                                       callee))

# }}}


# {{{ rename_callable

def rename_callable(
            t_unit: TranslationUnit,
            old_name: str,
            new_name: str | None = None,
            existing_ok=False
        ) -> TranslationUnit:
    """
    :arg program: An instance of :class:`loopy.TranslationUnit`
    :arg old_name: The callable to be renamed
    :arg new_name: New name for the callable to be renamed
    :arg existing_ok: An instance of :class:`bool`
    """
    from pymbolic import var

    from loopy.symbolic import (
        RuleAwareSubstitutionMapper,
        SubstitutionRuleMappingContext,
    )

    assert isinstance(t_unit, TranslationUnit)
    assert isinstance(old_name, str)

    if (new_name in t_unit.callables_table) and not existing_ok:
        raise LoopyError(f"callables named '{new_name}' already exists")

    if new_name is None:
        namegen = UniqueNameGenerator(
                    {n for n in t_unit.callables_table if isinstance(n, str)})
        new_name = namegen(old_name)

    assert isinstance(new_name, str)

    new_callables_table = {}

    for name, clbl in t_unit.callables_table.items():
        if name == old_name:
            name = new_name

        if isinstance(clbl, CallableKernel):
            knl = clbl.subkernel
            rule_mapping_context = SubstitutionRuleMappingContext(
                    knl.substitutions, knl.get_var_name_generator())
            smap = RuleAwareSubstitutionMapper(rule_mapping_context,
                                               {var(old_name): var(new_name)}.get,
                                               within=lambda *args: True)
            knl = rule_mapping_context.finish_kernel(smap.map_kernel(knl))
            clbl = clbl.copy(subkernel=knl.copy(name=name))
        elif isinstance(clbl, ScalarCallable):
            pass
        else:
            raise NotImplementedError(f"{type(clbl)}")

        new_callables_table[name] = clbl

    new_entrypoints = t_unit.entrypoints.copy()
    if old_name in new_entrypoints:
        new_entrypoints = ((new_entrypoints | frozenset([new_name]))
                           - frozenset([old_name]))

    return t_unit.copy(callables_table=constantdict(new_callables_table),
                        entrypoints=new_entrypoints)

# }}}

# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2017 Nick Curtis"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import ctypes
import logging
import os
import tempfile
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Sequence

import numpy as np
from codepy.jit import compile_from_string
from codepy.toolchain import GCCToolchain, ToolchainGuessError, guess_toolchain

from pytools import memoize_method
from pytools.codegen import CodeGenerator, Indentation
from pytools.prefork import ExecError

from loopy.kernel.array import ArrayBase
from loopy.target.execution import (
    ExecutionWrapperGeneratorBase,
    ExecutorBase,
    get_highlighted_code,
)
from loopy.types import LoopyType


if TYPE_CHECKING:
    from constantdict import constantdict

    from loopy.codegen.result import GeneratedProgram
    from loopy.kernel import LoopKernel
    from loopy.kernel.data import ArrayArg
    from loopy.schedule.tools import KernelArgInfo
    from loopy.translation_unit import TranslationUnit
    from loopy.typing import Expression


logger = logging.getLogger(__name__)

DEF_EVEN_DIV_FUNCTION = """
def _lpy_even_div(a, b):
    result, remdr = divmod(a, b)
    if remdr != 0:
        # FIXME: This error message is kind of crummy.
        raise ValueError("expected even division")
    return result


def _lpy_even_div_none(a, b):
    if a is None:
        return None

    result, remdr = divmod(a, b)
    if remdr != 0:
        # FIXME: This error message is kind of crummy.
        raise ValueError("expected even division")
    return result
"""


# {{{ CExecutionWrapperGenerator

class CExecutionWrapperGenerator(ExecutionWrapperGeneratorBase):
    """
    Specialized form of the :class:`ExecutionWrapperGeneratorBase` for
    pyopencl execution
    """

    def __init__(self):
        system_args = ["_lpy_c_kernels"]
        super().__init__(system_args)

    def python_dtype_str_inner(self, dtype):
        if np.dtype(str(dtype)).isbuiltin:
            name = dtype.name
            if dtype.name == "bool":
                name = "bool_"
            return f"_lpy_np.dtype(_lpy_np.{name})"
        raise Exception(f"dtype: {dtype} not recognized")

    # {{{ handle non numpy arguments

    def handle_non_numpy_arg(self, gen, arg):
        pass

    # }}}

    # {{{ handle allocation of unspecified arguments

    def handle_alloc(
            self, gen: CodeGenerator, arg: ArrayArg,
            strify: Callable[[Expression | tuple[Expression]], str],
            skip_arg_checks: bool) -> None:
        """
        Handle allocation of non-specified arguments for C-execution
        """
        from pymbolic import var

        assert isinstance(arg.shape, tuple)
        assert isinstance(arg.dtype, LoopyType)
        num_axes = len(arg.shape)
        for i in range(num_axes):
            gen("_lpy_shape_%d = %s" % (i, strify(arg.shape[i])))

        from loopy.kernel.array import get_strides
        strides = get_strides(arg)
        num_axes = len(strides)

        itemsize = arg.dtype.numpy_dtype.itemsize
        for i in range(num_axes):
            gen("_lpy_strides_%d = %s" % (i, strify(
                itemsize*strides[i])))

        if not skip_arg_checks:
            for i in range(num_axes):
                gen("assert _lpy_strides_%d > 0, "
                        "\"'%s' has negative stride in axis %d\""
                        % (i, arg.name, i))

        sym_strides = tuple(
                var("_lpy_strides_%d" % i)
                for i in range(num_axes))

        sym_shape = tuple(
                var("_lpy_shape_%d" % i)
                for i in range(num_axes))

        # find order of array
        from loopy.kernel.array import get_strides
        strides = get_strides(arg)
        order = "'C'" if (arg.shape == () or strides[-1] == 1) else "'F'"

        gen(f"{arg.name} = _lpy_np.empty({strify(sym_shape)}, "
                f"{self.python_dtype_str(gen, arg.dtype.numpy_dtype)}, "
                f"order={order})")

        expected_strides = tuple(
                var("_lpy_expected_strides_%s" % i)
                for i in range(num_axes))

        gen("{} = {}.strides".format(strify(expected_strides), arg.name))

        # check strides
        if not skip_arg_checks:
            strides_check_expr = self.get_strides_check_expr(
                    [strify(s) for s in sym_shape],
                    [strify(s) for s in sym_strides],
                    [strify(s) for s in expected_strides])
            gen(f"assert {strides_check_expr}, "
                    f"'Strides of loopy created array {arg.name}, "
                    "do not match expected.'")
            for i in range(num_axes):
                gen("del _lpy_shape_%d" % i)
                gen("del _lpy_strides_%d" % i)
            gen("")

    # }}}

    def target_specific_preamble(self, gen):
        """
        Add default C-imports to preamble
        """
        gen.add_to_preamble("import numpy as _lpy_np")
        gen.add_to_preamble(DEF_EVEN_DIV_FUNCTION)

    def initialize_system_args(self, gen):
        """
        Initializes possibly empty system arguments
        """
        pass

    # {{{ generate invocation

    def generate_invocation(self, gen: CodeGenerator, kernel: LoopKernel,
            kai: KernelArgInfo, host_program_name: str, args: Sequence[str]) -> None:
        gen("for knl in _lpy_c_kernels:")
        with Indentation(gen):
            gen(f"knl({', '.join(args)})")

    # }}}

    # {{{

    def generate_output_handler(self, gen: CodeGenerator,
            kernel: LoopKernel, kai: KernelArgInfo) -> None:
        options = kernel.options

        if options.return_dict:
            gen("return None, {%s}"
                    % ", ".join(f'"{arg_name}": {arg_name}'
                        for arg_name in kai.passed_arg_names
                        if kernel.arg_dict[arg_name].is_output))
        else:
            out_names = [arg_name for arg_name in kai.passed_arg_names
                    if kernel.arg_dict[arg_name].is_output]
            if out_names:
                gen(f"return None, ({', '.join(out_names)},)")
            else:
                gen("return None, ()")

    # }}}

    def generate_host_code(self, gen, codegen_result):
        # "host" code for C is embedded in the same file as the "device" code
        # this will enable a logical jumping off point for global barriers for
        # OpenMP, etc.
        pass

    def get_arg_pass(self, arg):
        return arg.name

# }}}


# {{{ CCompiler

class CCompiler:
    """
    The compiler module handles invocation of compilers to generate a shared lib
    using codepy, which can subsequently be loaded via ctypes.

    The general strategy here is as follows:

    1.  A :class:`codepy.Toolchain` is guessed from distutils.
        The user may override any flags obtained therein by passing in arguments
        to cc, cflags, etc.

    2.  The kernel source is built into and object first, then made into a shared
        library using :meth:`codepy.jit.compile_from_string`, which additionally
        handles caching

    3.  The resulting shared library is turned into a :class:`ctypes.CDLL`
        to enable calling by the invoker generated by, e.g.,
        :class:`CExecutionWrapperGenerator`
    """

    def __init__(self, toolchain=None,
                 cc="gcc", cflags=None,
                 ldflags=None, libraries=None,
                 include_dirs=None, library_dirs=None, defines=None,
                 source_suffix="c"):
        if cflags is None:
            cflags = "-std=c99 -O3 -fPIC".split()
        if ldflags is None:
            ldflags = "-shared".split()
        if libraries is None:
            libraries = []
        if include_dirs is None:
            include_dirs = []
        if library_dirs is None:
            library_dirs = []
        if defines is None:
            defines = []

        # try to get a default toolchain
        # or subclass supplied version if available
        self.toolchain = toolchain
        if toolchain is None:
            try:
                self.toolchain = guess_toolchain()
            except (ToolchainGuessError, ExecError):
                # missing compiler python was built with (likely, Conda)
                # use a default GCCToolchain
                logger = logging.getLogger(__name__)
                logger.warn("Default toolchain guessed from python config "
                            "not found, replacing with default GCCToolchain.")
                # this is ugly, but I'm not sure there's a clean way to copy the
                # default args
                self.toolchain = GCCToolchain(
                    cc="gcc",
                    ld="ld",
                    cflags="-std=c99 -O3 -fPIC".split(),
                    ldflags="-shared".split(),
                    libraries=[],
                    library_dirs=[],
                    defines=[],
                    undefines=[],
                    so_ext=".so",
                    o_ext=".o",
                    include_dirs=[],
                    features=set())

        if toolchain is None:
            # copy in all differing values
            diff = {"cc": cc,
                    "cflags": cflags,
                    "ldflags": ldflags,
                    "libraries": libraries,
                    "include_dirs": include_dirs,
                    "library_dirs": library_dirs,
                    "defines": defines}
            # filter empty and those equal to toolchain defaults
            diff = {k: v for k, v in diff.items()
                    if v and (not hasattr(self.toolchain, k) or
                              getattr(self.toolchain, k) != v)}
            self.toolchain = self.toolchain.copy(**diff)
        self.tempdir = tempfile.mkdtemp(prefix="tmp_loopy")
        self.source_suffix = source_suffix

    def _tempname(self, name):
        """Build temporary filename path in tempdir."""
        return os.path.join(self.tempdir, name)

    def build(self, name, code, debug=False, wait_on_error=None,
              debug_recompile=True, extra_build_options: Sequence[str] = ()):
        """Compile code, build and load shared library."""
        logger.debug(code)
        c_fname = self._tempname("code." + self.source_suffix)

        # build object
        _, _mod_name, ext_file, recompiled = \
            compile_from_string(
                self.toolchain.copy(
                    cflags=[*self.toolchain.cflags, *extra_build_options]),
                name, code,
                source_name=c_fname,
                cache_dir=self.tempdir,
                debug=debug,
                debug_recompile=debug_recompile,
                object=False)

        if recompiled:
            logger.debug(f"Kernel {name} compiled from source")
        else:
            logger.debug(f"Kernel {name} retrieved from cache")

        # and return compiled
        return ctypes.CDLL(ext_file)

# }}}


# {{{ CPlusPlusCompiler

class CPlusPlusCompiler(CCompiler):
    """Subclass of CCompiler to invoke a C++ compiler."""

    def __init__(self, toolchain=None,
                 cc="g++", cflags=None,
                 ldflags=None, libraries=None,
                 include_dirs=None, library_dirs=None, defines=None,
                 source_suffix="cpp"):

        super().__init__(
            toolchain=toolchain, cc=cc, cflags=cflags, ldflags=ldflags,
            libraries=libraries, include_dirs=include_dirs,
            library_dirs=library_dirs, defines=defines, source_suffix=source_suffix)

# }}}


# {{{ placeholder till ctypes fixes: https://github.com/python/cpython/issues/61103

class Complex64(ctypes.Structure):
    _fields_: ClassVar = [("real", ctypes.c_float), ("imag", ctypes.c_float)]


class Complex128(ctypes.Structure):
    _fields_: ClassVar = [("real", ctypes.c_double), ("imag", ctypes.c_double)]


class Complex256(ctypes.Structure):
    _fields_: ClassVar = [("real", ctypes.c_longdouble), ("imag", ctypes.c_longdouble)]


_NUMPY_COMPLEX_TYPE_TO_CTYPE = {
        np.complex64: Complex64,
        np.complex128: Complex128,
        }
if hasattr(np, "complex256"):
    _NUMPY_COMPLEX_TYPE_TO_CTYPE[np.complex256] = Complex256

# }}}


# {{{ _args_to_ctypes

def _args_to_ctypes(kernel: LoopKernel, passed_names: Sequence[str]):
    def _dtype_to_ctype(dtype):
        """Map NumPy dtype to equivalent ctypes type."""
        if dtype.is_complex():
            # complex ctypes aren't exposed
            np_dtype = dtype.numpy_dtype.type
            basetype = _NUMPY_COMPLEX_TYPE_TO_CTYPE[np_dtype]
        else:
            basetype = np.ctypeslib.as_ctypes_type(dtype)
        return basetype

    arg_info = []
    for arg_name in passed_names:
        arg = kernel.arg_dict[arg_name]

        ctype = _dtype_to_ctype(arg.dtype)
        if isinstance(arg, ArrayBase):
            ctype = ctypes.POINTER(ctype)
        arg_info.append(ctype)

    return arg_info

# }}}


# {{{ CompiledCKernel

class CompiledCKernel:
    """
    A CompiledCKernel wraps a loopy kernel, compiling it and loading the
    result as a shared library, and provides access to the kernel as a
    ctypes function object, wrapped by the __call__ method, which attempts
    to automatically map argument types.
    """

    def __init__(self, kernel: LoopKernel, devprog: GeneratedProgram,
            passed_names: Sequence[str], dev_code: str,
            comp: CCompiler | None = None):
        # get code and build
        self.code = dev_code
        self.comp = comp if comp is not None else CCompiler()
        self.dll = self.comp.build(devprog.name, self.code,
                                   extra_build_options=kernel.options.build_options)

        # get the function declaration for interface with ctypes
        self._fn = getattr(self.dll, devprog.name)
        # kernels are void by defn.
        self._fn.restype = None
        self._fn.argtypes = _args_to_ctypes(kernel, passed_names)

    def __call__(self, *args):
        """Execute kernel with given args mapped to ctypes equivalents."""
        args_ = []
        for arg, arg_t in zip(args, self._fn.argtypes):
            if hasattr(arg, "ctypes"):
                if arg.size == 0:
                    # TODO eliminate unused arguments from kernel
                    arg_ = arg_t(0.0)
                else:
                    arg_ = arg.ctypes.data_as(arg_t)
            else:
                arg_ = arg_t(arg)
            args_.append(arg_)
        self._fn(*args_)

# }}}


@dataclass(frozen=True)
class _KernelInfo:
    t_unit: TranslationUnit
    c_kernels: Sequence[CompiledCKernel]
    invoker: Callable[..., Any]


# {{{ CExecutor

class CExecutor(ExecutorBase):
    """An object connecting a kernel to a :class:`CompiledKernel`
    for execution.

    .. automethod:: __init__
    .. automethod:: __call__
    """

    def __init__(self, program, entrypoint, compiler: CCompiler | None = None):
        """
        :arg kernel: may be a loopy.LoopKernel, a generator returning kernels
            (a warning will be issued if more than one is returned). If the
            kernel has not yet been loop-scheduled, that is done, too, with no
            specific arguments.
        """

        self.compiler = compiler if compiler else CCompiler()
        super().__init__(program, entrypoint)

    def get_invoker_uncached(self, kernel, entrypoint, codegen_result):
        generator = CExecutionWrapperGenerator()
        return generator(kernel, entrypoint, codegen_result)

    def get_wrapper_generator(self):
        return CExecutionWrapperGenerator()

    @memoize_method
    def translation_unit_info(self,
            arg_to_dtype: constantdict[str, LoopyType] | None = None) -> _KernelInfo:
        t_unit = self.get_typed_and_scheduled_translation_unit(arg_to_dtype)

        from loopy.codegen import generate_code_v2
        codegen_result = generate_code_v2(t_unit)

        dev_code = codegen_result.device_code()
        host_code = codegen_result.host_code()
        all_code = "\n".join([dev_code, "", host_code])

        if t_unit[self.entrypoint].options.write_code:
            output = all_code
            if t_unit[self.entrypoint].options.allow_terminal_colors:
                output = get_highlighted_code(output)

            if t_unit[self.entrypoint].options.write_code is True:
                print(output)
            else:
                with open(t_unit[self.entrypoint].options.write_code, "w") as outf:
                    outf.write(output)

        if t_unit[self.entrypoint].options.edit_code:
            from pytools import invoke_editor
            dev_code = invoke_editor(dev_code, "code.c")
            # update code from editor
            all_code = "\n".join([dev_code, "", host_code])

        c_kernels = []

        from loopy.schedule.tools import get_kernel_arg_info
        kai = get_kernel_arg_info(t_unit[self.entrypoint])
        for dp in codegen_result.device_programs:
            c_kernels.append(CompiledCKernel(
                t_unit[self.entrypoint], dp, kai.passed_names, all_code,
                self.compiler))

        return _KernelInfo(
                t_unit=t_unit,
                c_kernels=c_kernels,
                invoker=self.get_invoker(t_unit, self.entrypoint, codegen_result))

    def __call__(self, *args, **kwargs):
        """
        :returns: ``(None, output)`` the output is a tuple of output arguments
            (arguments that are written as part of the kernel). The order is given
            by the order of kernel arguments. If this order is unspecified
            (such as when kernel arguments are inferred automatically),
            enable :attr:`loopy.Options.return_dict` to make *output* a
            :class:`dict` instead, with keys of argument names and values
            of the returned arrays.
        """
        if __debug__:
            self.check_for_required_array_arguments(kwargs.keys())

        if self.packing_controller is not None:
            kwargs = self.packing_controller(kwargs)

        program_info = self.translation_unit_info(self.arg_to_dtype(kwargs))

        return program_info.invoker(
                program_info.c_kernels, *args, **kwargs)

# }}}

# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = """
Copyright (C) 2015 James Stevens
Copyright (C) 2018 Kaushik Kulkarni
Copyright (C) 2019 Andreas Kloeckner
"""


__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from functools import cached_property, partial
from typing import TYPE_CHECKING, ClassVar

import islpy as isl
from islpy import dim_type
from pymbolic.mapper import CombineMapper
from pytools import ImmutableRecord, memoize_method

import loopy as lp
from loopy.diagnostic import LoopyError, warn_with_kernel
from loopy.kernel.data import AddressSpace, MultiAssignmentBase, TemporaryVariable
from loopy.kernel.function_interface import CallableKernel
from loopy.symbolic import CoefficientCollector, flatten
from loopy.translation_unit import TranslationUnit


if TYPE_CHECKING:
    from collections.abc import Sequence


__doc__ = """

.. currentmodule:: loopy

.. autoclass:: ToCountMap
.. autoclass:: ToCountPolynomialMap
.. autoclass:: CountGranularity
.. autoclass:: Op
.. autoclass:: MemAccess

.. autofunction:: get_op_map
.. autofunction:: get_mem_access_map
.. autofunction:: get_synchronization_map

.. autofunction:: gather_access_footprints
.. autofunction:: gather_access_footprint_bytes

.. currentmodule:: loopy.statistics

.. autoclass:: GuardedPwQPolynomial

.. currentmodule:: loopy
"""


# FIXME:
# - The SUBGROUP granularity is completely broken if the root kernel
#   contains the grid and the operations get counted in the callee.
#   To test, most of those are set to WORKITEM instead below (marked
#   with FIXMEs). This leads to value mismatches and key errors in
#   the tests.
# - Currently, nothing prevents summation across different
#   granularities, which is guaranteed to yield bogus results.
# - AccessFootprintGatherer needs to be redone to match get_op_map and
#   get_mem_access_map style
# - Test for the subkernel functionality need to be written


def get_kernel_parameter_space(kernel):
    return isl.Space.create_from_names(kernel.isl_context,
            set=[], params=sorted(kernel.outer_params())).params()


def get_kernel_zero_pwqpolynomial(kernel):
    space = get_kernel_parameter_space(kernel)
    space = space.insert_dims(dim_type.out, 0, 1)
    return isl.PwQPolynomial.zero(space)


# {{{ GuardedPwQPolynomial

def _get_param_tuple(obj):
    return tuple(
            obj.get_dim_name(dim_type.param, i)
            for i in range(obj.dim(dim_type.param)))


class GuardedPwQPolynomial:
    def __init__(self, pwqpolynomial, valid_domain):
        assert isinstance(pwqpolynomial, isl.PwQPolynomial)
        self.pwqpolynomial = pwqpolynomial
        self.valid_domain = valid_domain

        assert (_get_param_tuple(pwqpolynomial.space)
                == _get_param_tuple(valid_domain.space))

    @property
    def space(self):
        return self.valid_domain.space

    def __add__(self, other):
        if isinstance(other, GuardedPwQPolynomial):
            return GuardedPwQPolynomial(
                    self.pwqpolynomial + other.pwqpolynomial,
                    self.valid_domain & other.valid_domain)
        else:
            return GuardedPwQPolynomial(
                    self.pwqpolynomial + other,
                    self.valid_domain)

    __radd__ = __add__

    def __mul__(self, other):
        if isinstance(other, GuardedPwQPolynomial):
            return GuardedPwQPolynomial(
                    self.pwqpolynomial * other.pwqpolynomial,
                    self.valid_domain & other.valid_domain)
        else:
            return GuardedPwQPolynomial(
                    self.pwqpolynomial * other,
                    self.valid_domain)

    __rmul__ = __mul__

    def eval_with_dict(self, value_dict):
        space = self.pwqpolynomial.space
        pt = isl.Point.zero(space.params())

        for i in range(space.dim(dim_type.param)):
            par_name = space.get_dim_name(dim_type.param, i)
            pt = pt.set_coordinate_val(
                dim_type.param, i, value_dict[par_name])

        if not (isl.Set.from_point(pt) <= self.valid_domain):
            raise ValueError("evaluation point outside of domain of "
                    "definition of piecewise quasipolynomial")

        return self.pwqpolynomial.eval(pt).to_python()

    @staticmethod
    def zero():
        p = isl.PwQPolynomial("{ 0 }")
        return GuardedPwQPolynomial(p, isl.Set.universe(p.domain().space))

    def __str__(self):
        return str(self.pwqpolynomial)

    def __repr__(self):
        return "Guarded" + repr(self.pwqpolynomial)

# }}}


# {{{ ToCountMap

class ToCountMap:
    """A map from work descriptors like :class:`Op` and :class:`MemAccess`
    to any arithmetic type.

    .. automethod:: __getitem__
    .. automethod:: __str__
    .. automethod:: __repr__
    .. automethod:: __len__
    .. automethod:: get
    .. automethod:: items
    .. automethod:: keys
    .. automethod:: values

    .. automethod:: copy
    .. automethod:: with_set_attributes

    .. automethod:: filter_by
    .. automethod:: filter_by_func
    .. automethod:: group_by
    .. automethod:: to_bytes
    .. automethod:: sum

    """

    def __init__(self, count_map=None):
        if count_map is None:
            count_map = {}

        self.count_map = count_map

    def _zero(self):
        return 0

    def __add__(self, other):
        result = self.count_map.copy()
        for k, v in other.count_map.items():
            result[k] = self.count_map.get(k, 0) + v
        return self.copy(count_map=result)

    def __radd__(self, other):
        if other != 0:
            raise ValueError("ToCountMap: Attempted to add ToCountMap "
                                "to {} {}. ToCountMap may only be added to "
                                "0 and other ToCountMap objects."
                                .format(type(other), other))

        return self

    def __mul__(self, other):
        if isinstance(other, GuardedPwQPolynomial):
            return self.copy({
                index: other*value
                for index, value in self.count_map.items()})
        else:
            raise ValueError("ToCountMap: Attempted to multiply "
                                "ToCountMap by {} {}."
                                .format(type(other), other))

    __rmul__ = __mul__

    def __getitem__(self, index):
        return self.count_map[index]

    def __repr__(self):
        return repr(self.count_map)

    def __str__(self):
        return "\n".join(
                f"{k}: {v}"
                for k, v in sorted(self.count_map.items(),
                    key=lambda k: str(k)))

    def __len__(self):
        return len(self.count_map)

    def get(self, key, default=None):
        return self.count_map.get(key, default)

    def items(self):
        return self.count_map.items()

    def keys(self):
        return self.count_map.keys()

    def values(self):
        return self.count_map.values()

    def copy(self, count_map=None):
        if count_map is None:
            count_map = self.count_map

        return type(self)(count_map=count_map)

    def with_set_attributes(self, **kwargs):
        return self.copy(count_map={
            key.copy(**kwargs): val
            for key, val in self.count_map.items()})

    def filter_by(self, **kwargs):
        """Remove items without specified key fields.

        :arg kwargs: Keyword arguments matching fields in the keys of the
            :class:`ToCountMap`, each given a list of allowable values for that
            key field.

        :return: A :class:`ToCountMap` containing the subset of the items in
            the original :class:`ToCountMap` that match the field values
            passed.

        Example usage::

            # (first create loopy kernel and specify array data types)

            params = {"n": 512, "m": 256, "l": 128}
            mem_map = lp.get_mem_access_map(knl)
            filtered_map = mem_map.filter_by(direction=["load"],
                                             variable=["a","g"])
            tot_loads_a_g = filtered_map.eval_and_sum(params)

            # (now use these counts to, e.g., predict performance)

        """

        new_count_map = {}

        class _Sentinel:
            pass

        new_kwargs = {}
        for arg_field, allowable_vals in kwargs.items():
            if arg_field == "dtype":
                from loopy.types import to_loopy_type
                allowable_vals = [to_loopy_type(dtype) for dtype in allowable_vals]

            new_kwargs[arg_field] = allowable_vals

        for key, val in self.count_map.items():
            if all(getattr(key, arg_field, _Sentinel) in allowable_vals
                    for arg_field, allowable_vals in new_kwargs.items()):
                new_count_map[key] = val

        return self.copy(count_map=new_count_map)

    def filter_by_func(self, func):
        """Keep items that pass a test.

        :arg func: A function that takes a map key a parameter and returns a
            :class:`bool`.

        :arg: A :class:`ToCountMap` containing the subset of the items in the
            original :class:`ToCountMap` for which func(key) is true.

        Example usage::

            # (first create loopy kernel and specify array data types)

            params = {"n": 512, "m": 256, "l": 128}
            mem_map = lp.get_mem_access_map(knl)
            def filter_func(key):
                return key.lid_strides[0] > 1 and key.lid_strides[0] <= 4:

            filtered_map = mem_map.filter_by_func(filter_func)
            tot = filtered_map.eval_and_sum(params)

            # (now use these counts to, e.g., predict performance)

        """

        new_count_map = {}

        for self_key, self_val in self.count_map.items():
            if func(self_key):
                new_count_map[self_key] = self_val

        return self.copy(count_map=new_count_map)

    def group_by(self, *args):
        """Group map items together, distinguishing by only the key fields
        passed in args.

        :arg args: Zero or more :class:`str` fields of map keys.

        :return: A :class:`ToCountMap` containing the same total counts grouped
            together by new keys that only contain the fields specified in the
            arguments passed.

        Example usage::

            # (first create loopy kernel and specify array data types)

            params = {"n": 512, "m": 256, "l": 128}
            mem_map = get_mem_access_map(knl)
            grouped_map = mem_map.group_by("mtype", "dtype", "direction")

            f32_global_ld = grouped_map[MemAccess(mtype="global",
                                                  dtype=np.float32,
                                                  direction="load")
                                       ].eval_with_dict(params)
            f32_global_st = grouped_map[MemAccess(mtype="global",
                                                  dtype=np.float32,
                                                  direction="store")
                                       ].eval_with_dict(params)
            f32_local_ld = grouped_map[MemAccess(mtype="local",
                                                 dtype=np.float32,
                                                 direction="load")
                                      ].eval_with_dict(params)
            f32_local_st = grouped_map[MemAccess(mtype="local",
                                                 dtype=np.float32,
                                                 direction="store")
                                      ].eval_with_dict(params)

            op_map = get_op_map(knl)
            ops_dtype = op_map.group_by("dtype")

            f32ops = ops_dtype[Op(dtype=np.float32)].eval_with_dict(params)
            f64ops = ops_dtype[Op(dtype=np.float64)].eval_with_dict(params)
            i32ops = ops_dtype[Op(dtype=np.int32)].eval_with_dict(params)

            # (now use these counts to, e.g., predict performance)

        """

        new_count_map = {}

        # make sure all item keys have same type
        if self.count_map:
            key_type = type(next(iter(self.keys())))
            if not all(isinstance(x, key_type) for x in self.keys()):
                raise ValueError("ToCountMap: group_by() function may only "
                                 "be used on ToCountMaps with uniform keys")
        else:
            return self

        for self_key, self_val in self.count_map.items():
            new_key = key_type(
                    **{
                        field: getattr(self_key, field)
                        for field in args})

            new_count_map[new_key] = new_count_map.get(new_key, 0) + self_val

        return self.copy(count_map=new_count_map)

    def to_bytes(self):
        """Convert counts to bytes using data type in map key.

        :return: A :class:`ToCountMap` mapping each original key to an
            :class:`islpy.PwQPolynomial` with counts in bytes rather than
            instances.

        Example usage::

            # (first create loopy kernel and specify array data types)

            bytes_map = get_mem_access_map(knl).to_bytes()
            params = {"n": 512, "m": 256, "l": 128}

            s1_g_ld_bytes = bytes_map.filter_by(
                                mtype=["global"], lid_strides={0: 1},
                                direction=["load"]).eval_and_sum(params)
            s2_g_ld_bytes = bytes_map.filter_by(
                                mtype=["global"], lid_strides={0: 2},
                                direction=["load"]).eval_and_sum(params)
            s1_g_st_bytes = bytes_map.filter_by(
                                mtype=["global"], lid_strides={0: 1},
                                direction=["store"]).eval_and_sum(params)
            s2_g_st_bytes = bytes_map.filter_by(
                                mtype=["global"], lid_strides={0: 2},
                                direction=["store"]).eval_and_sum(params)

            # (now use these counts to, e.g., predict performance)

        """

        new_count_map = {}

        for key, val in self.count_map.items():
            new_count_map[key] = int(key.dtype.itemsize) * val

        return self.copy(new_count_map)

    def sum(self):
        """:return: A sum of the values of the dictionary."""

        total = self._zero()

        for v in self.count_map.values():
            total = v + total

        return total

# }}}


# {{{ ToCountPolynomialMap

class ToCountPolynomialMap(ToCountMap):
    """Maps any type of key to a :class:`islpy.PwQPolynomial` or a
    :class:`~loopy.statistics.GuardedPwQPolynomial`.

    .. automethod:: eval_and_sum
    """

    def __init__(self, space, count_map=None):
        if not isinstance(space, isl.Space):
            raise TypeError(
                    "first argument to ToCountPolynomialMap must be "
                    "of type islpy.Space")

        assert space.is_params()
        self.space = space

        space_param_tuple = _get_param_tuple(space)

        for val in count_map.values():
            if isinstance(val, isl.PwQPolynomial):
                assert val.dim(dim_type.out) == 1
            elif isinstance(val, GuardedPwQPolynomial):
                assert val.pwqpolynomial.dim(dim_type.out) == 1
            else:
                raise TypeError("unexpected value type")

            assert _get_param_tuple(val.space) == space_param_tuple

        super().__init__(count_map)

    def _zero(self):
        space = self.space.insert_dims(dim_type.out, 0, 1)
        return isl.PwQPolynomial.zero(space)

    def copy(self, count_map=None, space=None):
        if count_map is None:
            count_map = self.count_map

        if space is None:
            space = self.space

        return type(self)(space, count_map)

    def eval_and_sum(self, params=None):
        """Add all counts and evaluate with provided parameter dict *params*

        :return: An :class:`int` containing the sum of all counts
            evaluated with the parameters provided.

        Example usage::

            # (first create loopy kernel and specify array data types)

            params = {"n": 512, "m": 256, "l": 128}
            mem_map = lp.get_mem_access_map(knl)
            filtered_map = mem_map.filter_by(direction=["load"],
                                             variable=["a", "g"])
            tot_loads_a_g = filtered_map.eval_and_sum(params)

            # (now use these counts to, e.g., predict performance)

        """
        if params is None:
            params = {}

        return self.sum().eval_with_dict(params)

# }}}


# {{{ subst_into_to_count_map

def subst_into_guarded_pwqpolynomial(new_space, guarded_poly, subst_dict):
    from loopy.isl_helpers import get_param_subst_domain, subst_into_pwqpolynomial

    poly = subst_into_pwqpolynomial(
            new_space, guarded_poly.pwqpolynomial, subst_dict)

    valid_domain = guarded_poly.valid_domain
    i_begin_subst_space = valid_domain.dim(dim_type.param)

    valid_domain, subst_domain, _ = get_param_subst_domain(
            new_space, guarded_poly.valid_domain, subst_dict)

    valid_domain = valid_domain & subst_domain
    valid_domain = valid_domain.project_out(dim_type.param, 0, i_begin_subst_space)
    return GuardedPwQPolynomial(poly, valid_domain)


def subst_into_to_count_map(space, tcm, subst_dict):
    from loopy.isl_helpers import subst_into_pwqpolynomial
    new_count_map = {}
    for key, value in tcm.count_map.items():
        if isinstance(value, GuardedPwQPolynomial):
            new_count_map[key] = subst_into_guarded_pwqpolynomial(
                    space, value, subst_dict)

        elif isinstance(value, isl.PwQPolynomial):
            new_count_map[key] = subst_into_pwqpolynomial(space, value, subst_dict)

        elif isinstance(value, int):
            new_count_map[key] = value

        else:
            raise ValueError("unexpected value type")

    return tcm.copy(space=space, count_map=new_count_map)

# }}}


# {{{ CountGranularity

class CountGranularity:
    """Strings specifying whether an operation should be counted once per
    *work-item*, *sub-group*, or *work-group*.

    .. attribute:: WORKITEM

       A :class:`str` that specifies that an operation should be counted
       once per *work-item*.

    .. attribute:: SUBGROUP

       A :class:`str` that specifies that an operation should be counted
       once per *sub-group*.

    .. attribute:: WORKGROUP

       A :class:`str` that specifies that an operation should be counted
       once per *work-group*.

    """

    WORKITEM = "workitem"
    SUBGROUP = "subgroup"
    WORKGROUP = "workgroup"
    ALL: ClassVar[Sequence[str]] = [WORKITEM, SUBGROUP, WORKGROUP]

# }}}


# {{{ Op descriptor

class Op(ImmutableRecord):
    """A descriptor for a type of arithmetic operation.

    .. attribute:: dtype

       A :class:`loopy.types.LoopyType` or :class:`numpy.dtype` that specifies the
       data type operated on.

    .. attribute:: name

       A :class:`str` that specifies the kind of arithmetic operation as
       *add*, *mul*, *div*, *pow*, *shift*, *bw* (bitwise), etc.

    .. attribute:: count_granularity

       A :class:`str` that specifies whether this operation should be counted
       once per *work-item*, *sub-group*, or *work-group*. The granularities
       allowed can be found in :class:`CountGranularity`, and may be accessed,
       e.g., as ``CountGranularity.WORKITEM``. A work-item is a single instance
       of computation executing on a single processor (think "thread"), a
       collection of which may be grouped together into a work-group. Each
       work-group executes on a single compute unit with all work-items within
       the work-group sharing local memory. A sub-group is an
       implementation-dependent grouping of work-items within a work-group,
       analogous to an NVIDIA CUDA warp.

    .. attribute:: kernel_name

        A :class:`str` representing the kernel name where the operation occurred.

    """

    def __init__(self, dtype=None, name=None, count_granularity=None,
            kernel_name=None):
        if count_granularity not in [*CountGranularity.ALL, None]:
            raise ValueError("Op.__init__: count_granularity '%s' is "
                    "not allowed. count_granularity options: %s"
                    % (count_granularity, [*CountGranularity.ALL, None]))

        if dtype is not None:
            from loopy.types import to_loopy_type
            dtype = to_loopy_type(dtype)

        super().__init__(dtype=dtype, name=name,
                        count_granularity=count_granularity,
                        kernel_name=kernel_name)

    def __repr__(self):
        # Record.__repr__ overridden for consistent ordering and conciseness
        if self.kernel_name is not None:
            return (f"Op({self.dtype}, {self.name}, {self.count_granularity},"
                    f' "{self.kernel_name}")')
        else:
            return f"Op({self.dtype}, {self.name}, {self.count_granularity})"

# }}}


# {{{ MemAccess descriptor

class MemAccess(ImmutableRecord):
    """A descriptor for a type of memory access.

    .. attribute:: mtype

       A :class:`str` that specifies the memory type accessed as **global**
       or **local**

    .. attribute:: dtype

       A :class:`loopy.types.LoopyType` or :class:`numpy.dtype` that specifies the
       data type accessed.

    .. attribute:: lid_strides

       A :class:`dict` of **{** :class:`int` **:**
       :data:`~pymbolic.typing.Expression` or :class:`int` **}** that
       specifies local strides for each local id in the memory access index.
       Local ids not found will not be present in ``lid_strides.keys()``.
       Uniform access (i.e. work-items within a sub-group access the same
       item) is indicated by setting ``lid_strides[0]=0``, but may also occur
       when no local id 0 is found, in which case the 0 key will not be
       present in lid_strides.

    .. attribute:: gid_strides

       A :class:`dict` of **{** :class:`int` **:**
       :data:`~pymbolic.typing.Expression` or :class:`int` **}** that
       specifies global strides for each global id in the memory access index.
       global ids not found will not be present in ``gid_strides.keys()``.

    .. attribute:: direction

       A :class:`str` that specifies the direction of memory access as
       **load** or **store**.

    .. attribute:: variable

       A :class:`str` that specifies the variable name of the data
       accessed.

    .. attribute:: variable_tags

       A :class:`frozenset` of subclasses of :class:`~pytools.tag.Tag`
       that reflects :attr:`~loopy.TaggedVariable.tags` of
       an accessed variable.

    .. attribute:: count_granularity

       A :class:`str` that specifies whether this operation should be counted
       once per *work-item*, *sub-group*, or *work-group*. The granularities
       allowed can be found in :class:`CountGranularity`, and may be accessed,
       e.g., as ``CountGranularity.WORKITEM``. A work-item is a single instance
       of computation executing on a single processor (think "thread"), a
       collection of which may be grouped together into a work-group. Each
       work-group executes on a single compute unit with all work-items within
       the work-group sharing local memory. A sub-group is an
       implementation-dependent grouping of work-items within a work-group,
       analogous to an NVIDIA CUDA warp.

    .. attribute:: kernel_name

        A :class:`str` representing the kernel name where the operation occurred.
    """

    def __init__(self, mtype=None, dtype=None, lid_strides=None, gid_strides=None,
                 direction=None, variable=None,
                 *, variable_tags=None,
                 count_granularity=None, kernel_name=None):

        if count_granularity not in [*CountGranularity.ALL, None]:
            raise ValueError("Op.__init__: count_granularity '%s' is "
                    "not allowed. count_granularity options: %s"
                    % (count_granularity, [*CountGranularity.ALL, None]))

        if variable_tags is None:
            variable_tags = frozenset()

        if dtype is not None:
            from loopy.types import to_loopy_type
            dtype = to_loopy_type(dtype)

        ImmutableRecord.__init__(self, mtype=mtype, dtype=dtype,
                            lid_strides=lid_strides,
                            gid_strides=gid_strides, direction=direction,
                            variable=variable, variable_tags=variable_tags,
                            count_granularity=count_granularity,
                            kernel_name=kernel_name)

    def __hash__(self):
        # dicts in gid_strides and lid_strides aren't natively hashable
        return hash(repr(self))

    def __repr__(self):
        # Record.__repr__ overridden for consistent ordering and conciseness
        return "MemAccess({}, {}, {}, {}, {}, {}, {}, {}, {})".format(
            self.mtype,
            self.dtype,
            None if self.lid_strides is None else dict(
                sorted(self.lid_strides.items())),
            None if self.gid_strides is None else dict(
                sorted(self.gid_strides.items())),
            self.direction,
            self.variable,
            "None" if not self.variable_tags else str(self.variable_tags),
            self.count_granularity,
            repr(self.kernel_name))

# }}}


# {{{ Sync descriptor

class Sync(ImmutableRecord):
    """A descriptor for a type of synchronization.

    .. attribute:: kind

       A string describing the synchronization kind, e.g. ``"barrier_global"`` or
       ``"barrier_local"`` or ``"kernel_launch"``.

    .. attribute:: kernel_name

        A :class:`str` representing the kernel name where the operation occurred.
    """

    def __init__(self, kind=None, kernel_name=None):
        super().__init__(kind=kind, kernel_name=kernel_name)

    def __repr__(self):
        # Record.__repr__ overridden for consistent ordering and conciseness
        return f"Sync({self.kind}, {self.kernel_name})"

# }}}


# {{{ CounterBase

class CounterBase(CombineMapper):
    def __init__(self, knl, callables_table, kernel_rec):
        self.knl = knl
        self.callables_table = callables_table
        self.kernel_rec = kernel_rec

        from loopy.type_inference import TypeReader
        self.type_inf = TypeReader(knl, callables_table)
        self.zero = get_kernel_zero_pwqpolynomial(self.knl)
        self.one = self.zero + 1

    @cached_property
    def param_space(self):
        return get_kernel_parameter_space(self.knl)

    def new_poly_map(self, count_map):
        return ToCountPolynomialMap(self.param_space, count_map)

    def new_zero_poly_map(self):
        return self.new_poly_map({})

    def combine(self, values):
        return sum(values)

    def map_constant(self, expr):
        return self.new_zero_poly_map()

    def map_call(self, expr):
        from loopy.symbolic import ResolvedFunction
        assert isinstance(expr.function, ResolvedFunction)
        clbl = self.callables_table[expr.function.name]

        from loopy.kernel.data import ValueArg
        from loopy.kernel.function_interface import (
            CallableKernel,
            get_kw_pos_association,
        )
        if isinstance(clbl, CallableKernel):
            sub_result = self.kernel_rec(clbl.subkernel)
            _, pos_to_kw = get_kw_pos_association(clbl.subkernel)

            subst_dict = {
                    pos_to_kw[i]: param
                    for i, param in enumerate(expr.parameters)
                    if isinstance(clbl.subkernel.arg_dict[pos_to_kw[i]],
                                  ValueArg)}

            return subst_into_to_count_map(
                    self.param_space,
                    sub_result, subst_dict) \
                    + self.rec(expr.parameters)

        else:
            raise NotImplementedError()

    def map_call_with_kwargs(self, expr):
        # See https://github.com/inducer/loopy/pull/323
        raise NotImplementedError

    def map_sum(self, expr):
        if expr.children:
            return sum(self.rec(child) for child in expr.children)
        else:
            return self.new_zero_poly_map()

    map_product = map_sum

    def map_comparison(self, expr):
        return self.rec(expr.left)+self.rec(expr.right)

    def map_if(self, expr):
        warn_with_kernel(self.knl, "summing_if_branches",
                         "%s counting sum of if-expression branches."
                         % type(self).__name__)
        return self.rec(expr.condition) + self.rec(expr.then) \
               + self.rec(expr.else_)

    def map_if_positive(self, expr):
        warn_with_kernel(self.knl, "summing_if_branches",
                         "%s counting sum of if-expression branches."
                         % type(self).__name__)
        return self.rec(expr.criterion) + self.rec(expr.then) \
               + self.rec(expr.else_)

    def map_common_subexpression(self, expr):
        raise RuntimeError("%s encountered %s--not supposed to happen"
                % (type(self).__name__, type(expr).__name__))

    map_derivative = map_common_subexpression
    map_slice = map_common_subexpression

    def map_reduction(self, expr):
        # preprocessing should have removed these
        raise RuntimeError("%s encountered %s--not supposed to happen"
                % (type(self).__name__, type(expr).__name__))

# }}}


# {{{ ExpressionOpCounter

class ExpressionOpCounter(CounterBase):
    def __init__(self, knl, callables_table, kernel_rec,
            count_within_subscripts=True):
        super().__init__(
                knl, callables_table, kernel_rec)
        self.count_within_subscripts = count_within_subscripts

    arithmetic_count_granularity = CountGranularity.SUBGROUP

    def combine(self, values):
        return sum(values)

    def map_constant(self, expr):
        return self.new_zero_poly_map()

    map_tagged_variable = map_constant
    map_variable = map_constant
    map_nan = map_constant

    def map_call(self, expr):
        from loopy.symbolic import ResolvedFunction
        assert isinstance(expr.function, ResolvedFunction)
        clbl = self.callables_table[expr.function.name]

        from loopy.kernel.function_interface import CallableKernel
        if not isinstance(clbl, CallableKernel):
            return self.new_poly_map(
                        {Op(dtype=self.type_inf(expr),
                            name="func:"+clbl.name,
                            count_granularity=self.arithmetic_count_granularity,
                            kernel_name=self.knl.name): self.one}
                        ) + self.rec(expr.parameters)
        else:
            return super().map_call(expr)

    def map_subscript(self, expr):
        if self.count_within_subscripts:
            return self.rec(expr.index)
        else:
            return self.new_zero_poly_map()

    def map_sub_array_ref(self, expr):
        # generates an array view, considered free
        return self.new_zero_poly_map()

    def map_sum(self, expr):
        assert expr.children
        return self.new_poly_map(
                    {Op(dtype=self.type_inf(expr),
                        name="add",
                        count_granularity=self.arithmetic_count_granularity,
                        kernel_name=self.knl.name):
                     self.zero + (len(expr.children)-1)}
                    ) + sum(self.rec(child) for child in expr.children)

    def map_product(self, expr):
        from pymbolic.primitives import is_zero
        assert expr.children
        return sum(self.new_poly_map({Op(dtype=self.type_inf(expr),
                                  name="mul",
                                  count_granularity=(
                                      self.arithmetic_count_granularity),
                                  kernel_name=self.knl.name): self.one})
                   + self.rec(child)
                   for child in expr.children
                   if not is_zero(child + 1)) + \
                   self.new_poly_map({Op(dtype=self.type_inf(expr),
                                  name="mul",
                                  count_granularity=(
                                      self.arithmetic_count_granularity),
                                  kernel_name=self.knl.name): -self.one})

    def map_quotient(self, expr, *args):
        return self.new_poly_map({Op(dtype=self.type_inf(expr),
                              name="div",
                              count_granularity=self.arithmetic_count_granularity,
                              kernel_name=self.knl.name): self.one}) \
                                + self.rec(expr.numerator) \
                                + self.rec(expr.denominator)

    map_floor_div = map_quotient
    map_remainder = map_quotient

    def map_power(self, expr):
        return self.new_poly_map({Op(dtype=self.type_inf(expr),
                              name="pow",
                              count_granularity=self.arithmetic_count_granularity,
                              kernel_name=self.knl.name): self.one}) \
                                + self.rec(expr.base) \
                                + self.rec(expr.exponent)

    def map_type_cast(self, expr):
        # Treats type casting as free
        return self.rec(expr.child)

    def map_left_shift(self, expr):
        return self.new_poly_map({Op(dtype=self.type_inf(expr),
                              name="shift",
                              count_granularity=self.arithmetic_count_granularity,
                              kernel_name=self.knl.name): self.one}) \
                                + self.rec(expr.shiftee) \
                                + self.rec(expr.shift)

    map_right_shift = map_left_shift

    def map_bitwise_not(self, expr):
        return self.new_poly_map({Op(dtype=self.type_inf(expr),
                              name="bw",
                              count_granularity=self.arithmetic_count_granularity,
                              kernel_name=self.knl.name): self.one}) \
                                + self.rec(expr.child)

    def map_bitwise_or(self, expr):
        return self.new_poly_map({Op(dtype=self.type_inf(expr),
                              name="bw",
                              count_granularity=self.arithmetic_count_granularity,
                              kernel_name=self.knl.name):
                           self.zero + (len(expr.children)-1)}) \
                                + sum(self.rec(child) for child in expr.children)

    map_bitwise_xor = map_bitwise_or
    map_bitwise_and = map_bitwise_or

    def map_if(self, expr):
        warn_with_kernel(self.knl, "summing_if_branches_ops",
                         "ExpressionOpCounter counting ops as sum of "
                         "if-statement branches.")
        return self.rec(expr.condition) + self.rec(expr.then) \
               + self.rec(expr.else_)

    def map_if_positive(self, expr):
        warn_with_kernel(self.knl, "summing_ifpos_branches_ops",
                         "ExpressionOpCounter counting ops as sum of "
                         "if_pos-statement branches.")
        return self.rec(expr.criterion) + self.rec(expr.then) \
               + self.rec(expr.else_)

    def map_min(self, expr):
        return self.new_poly_map({Op(dtype=self.type_inf(expr),
                              name="maxmin",
                              count_granularity=self.arithmetic_count_granularity,
                              kernel_name=self.knl.name):
                           len(expr.children)-1}) \
               + sum(self.rec(child) for child in expr.children)

    map_max = map_min

    def map_common_subexpression(self, expr):
        raise NotImplementedError("ExpressionOpCounter encountered "
                                  "common_subexpression, "
                                  "map_common_subexpression not implemented.")

    def map_derivative(self, expr):
        raise NotImplementedError("ExpressionOpCounter encountered "
                                  "derivative, "
                                  "map_derivative not implemented.")

    def map_slice(self, expr):
        raise NotImplementedError("ExpressionOpCounter encountered slice, "
                                  "map_slice not implemented.")

# }}}


# {{{ modified coefficient collector that ignores denominator of floor div

class _IndexStrideCoefficientCollector(CoefficientCollector):

    def map_floor_div(self, expr):
        from warnings import warn
        warn("_IndexStrideCoefficientCollector encountered FloorDiv, ignoring "
             "denominator in expression %s" % (expr), stacklevel=1)
        return self.rec(expr.numerator)

# }}}


# {{{ _get_lid_and_gid_strides

def _get_lid_and_gid_strides(knl, array, index):
    # find all local and global index tags and corresponding inames
    from loopy.symbolic import get_dependencies
    my_inames = get_dependencies(index) & knl.all_inames()

    from loopy.kernel.data import (
        GroupInameTag,
        LocalInameTag,
        filter_iname_tags_by_type,
    )
    lid_to_iname = {}
    gid_to_iname = {}
    for iname in my_inames:
        tags = knl.iname_tags_of_type(iname, (GroupInameTag, LocalInameTag))
        if tags:
            tag, = filter_iname_tags_by_type(
                tags, (GroupInameTag, LocalInameTag), 1)
            if isinstance(tag, LocalInameTag):
                lid_to_iname[tag.axis] = iname
            else:
                gid_to_iname[tag.axis] = iname

    # create lid_strides and gid_strides dicts

    # strides are coefficients in flattened index, i.e., we want
    # lid_strides = {0:l0, 1:l1, 2:l2, ...} and
    # gid_strides = {0:g0, 1:g1, 2:g2, ...},
    # where l0, l1, l2, g0, g1, and g2 come from flattened index
    # [... + g2*gid2 + g1*gid1 + g0*gid0 + ... + l2*lid2 + l1*lid1 + l0*lid0]

    from pymbolic.primitives import Variable

    from loopy.diagnostic import ExpressionNotAffineError
    from loopy.kernel.array import FixedStrideArrayDimTag
    from loopy.symbolic import simplify_using_aff

    def get_iname_strides(tag_to_iname_dict):
        tag_to_stride_dict = {}

        if array.dim_tags is None:
            assert len(index) <= 1
            dim_tags = (None,) * len(index)
        else:
            dim_tags = array.dim_tags

        for tag in tag_to_iname_dict:
            total_iname_stride = 0
            # find total stride of this iname for each axis
            for idx, axis_tag in zip(index, dim_tags):
                # collect index coefficients
                try:
                    coeffs = _IndexStrideCoefficientCollector(
                            [tag_to_iname_dict[tag]])(
                                    simplify_using_aff(knl, idx))
                except ExpressionNotAffineError:
                    total_iname_stride = None
                    break

                # check if idx contains this iname
                try:
                    coeff = coeffs[Variable(tag_to_iname_dict[tag])]
                except KeyError:
                    # idx does not contain this iname
                    continue

                # found coefficient of this iname
                # now determine stride
                if isinstance(axis_tag, FixedStrideArrayDimTag):
                    axis_tag_stride = axis_tag.stride

                    if axis_tag_stride is lp.auto:
                        total_iname_stride = None
                        break

                elif axis_tag is None:
                    axis_tag_stride = 1

                else:
                    continue

                total_iname_stride += axis_tag_stride*coeff

            tag_to_stride_dict[tag] = flatten(total_iname_stride)

        return tag_to_stride_dict

    return get_iname_strides(lid_to_iname), get_iname_strides(gid_to_iname)

# }}}


# {{{ MemAccessCounterBase

class MemAccessCounterBase(CounterBase):
    def map_sub_array_ref(self, expr):
        # generates an array view, considered free
        return self.new_zero_poly_map()

    def map_call(self, expr):
        from loopy.symbolic import ResolvedFunction
        assert isinstance(expr.function, ResolvedFunction)
        clbl = self.callables_table[expr.function.name]

        from loopy.kernel.function_interface import CallableKernel
        if not isinstance(clbl, CallableKernel):
            return self.rec(expr.parameters)
        else:
            return super().map_call(expr)

# }}}


# {{{ LocalMemAccessCounter

class LocalMemAccessCounter(MemAccessCounterBase):
    local_mem_count_granularity = CountGranularity.SUBGROUP

    def count_var_access(self, dtype, name, index):
        count_map = {}
        if name in self.knl.temporary_variables:
            array = self.knl.temporary_variables[name]
            if isinstance(array, TemporaryVariable) and (
                    array.address_space == AddressSpace.LOCAL):
                if index is None:
                    # no subscript
                    count_map[MemAccess(
                                mtype="local",
                                dtype=dtype,
                                count_granularity=self.local_mem_count_granularity,
                                kernel_name=self.knl.name)] = self.one
                    return self.new_poly_map(count_map)

                array = self.knl.temporary_variables[name]

                # could be tuple or scalar index
                index_tuple = index
                if not isinstance(index_tuple, tuple):
                    index_tuple = (index_tuple,)

                lid_strides, gid_strides = _get_lid_and_gid_strides(
                                                self.knl, array, index_tuple)

                count_map[MemAccess(
                        mtype="local",
                        dtype=dtype,
                        lid_strides=dict(sorted(lid_strides.items())),
                        gid_strides=dict(sorted(gid_strides.items())),
                        variable=name,
                        count_granularity=self.local_mem_count_granularity,
                        kernel_name=self.knl.name)] = self.one

        return self.new_poly_map(count_map)

    def map_variable(self, expr):
        return self.count_var_access(
                    self.type_inf(expr), expr.name, None)

    map_tagged_variable = map_variable

    def map_subscript(self, expr):
        return (self.count_var_access(self.type_inf(expr),
                                      expr.aggregate.name,
                                      expr.index)
                + self.rec(expr.index))

# }}}


# {{{ GlobalMemAccessCounter

class GlobalMemAccessCounter(MemAccessCounterBase):
    def map_variable(self, expr):
        name = expr.name

        if name in self.knl.arg_dict:
            array = self.knl.arg_dict[name]
        else:
            # this is a temporary variable
            # FIXME temporary variable could have global address space
            return self.new_zero_poly_map()

        if not isinstance(array, lp.ArrayArg):
            # this array is not in global memory
            return self.new_zero_poly_map()

        return self.new_poly_map({MemAccess(mtype="global",
                    dtype=self.type_inf(expr), lid_strides={},
                    gid_strides={}, variable=name,
                    count_granularity=CountGranularity.WORKITEM,
                    kernel_name=self.knl.name): self.one}
                    ) + self.rec(expr.index)

    def map_subscript(self, expr):
        name = expr.aggregate.name
        try:
            var_tags = expr.aggregate.tags
        except AttributeError:
            var_tags = frozenset()

        is_global_temp = False
        if name in self.knl.arg_dict:
            array = self.knl.arg_dict[name]
        elif name in self.knl.temporary_variables:
            # This a temporary, but might have global address space
            from loopy.kernel.data import AddressSpace
            array = self.knl.temporary_variables[name]
            if array.address_space != AddressSpace.GLOBAL:
                # This temporary does not have global address space
                return self.rec(expr.index)
            # This temporary has global address space
            is_global_temp = True
        else:
            # This temporary does not have global address space
            return self.rec(expr.index)

        if (not is_global_temp) and not isinstance(array, lp.ArrayArg):
            # This array is not in global memory
            return self.rec(expr.index)

        index_tuple = expr.index  # could be tuple or scalar index
        if not isinstance(index_tuple, tuple):
            index_tuple = (index_tuple,)

        lid_strides, gid_strides = _get_lid_and_gid_strides(
                                        self.knl, array, index_tuple)

        global_access_count_granularity = CountGranularity.SUBGROUP

        # Account for broadcasts once per subgroup
        count_granularity = CountGranularity.WORKITEM if (
                # if the stride in lid.0 is known
                0 in lid_strides
                and
                # it is nonzero
                lid_strides[0] != 0
                ) else global_access_count_granularity

        return self.new_poly_map({MemAccess(
                            mtype="global",
                            dtype=self.type_inf(expr),
                            lid_strides=dict(sorted(lid_strides.items())),
                            gid_strides=dict(sorted(gid_strides.items())),
                            variable=name,
                            variable_tags=var_tags,
                            count_granularity=count_granularity,
                            kernel_name=self.knl.name,
                            ): self.one}
                          ) + self.rec(expr.index_tuple)

# }}}


# {{{ AccessFootprintGatherer

class AccessFootprintGatherer(CombineMapper):
    def __init__(self, kernel, domain, ignore_uncountable=False):
        self.kernel = kernel
        self.domain = domain
        self.ignore_uncountable = ignore_uncountable

    @staticmethod
    def combine(values):
        assert values

        def merge_dicts(a, b):
            result = a.copy()

            for var_name, footprint in b.items():
                if var_name in result:
                    result[var_name] = result[var_name] | footprint
                else:
                    result[var_name] = footprint

            return result

        from functools import reduce
        return reduce(merge_dicts, values)

    def map_constant(self, expr):
        return {}

    def map_variable(self, expr):
        return {}

    def map_subscript(self, expr):
        subscript = expr.index

        if not isinstance(subscript, tuple):
            subscript = (subscript,)

        from loopy.symbolic import get_access_map

        try:
            access_range = get_access_map(self.domain, subscript,
                    self.kernel.assumptions).range()
        except isl.Error as err:
            # Likely: index was non-linear, nothing we can do.
            if self.ignore_uncountable:
                return {}
            else:
                raise LoopyError("failed to gather footprint: %s" % expr) from err

        except TypeError as err:
            # Likely: index was non-linear, nothing we can do.
            if self.ignore_uncountable:
                return {}
            else:
                raise LoopyError("failed to gather footprint: %s" % expr) from err

        from pymbolic.primitives import Variable
        assert isinstance(expr.aggregate, Variable)

        return self.combine([
            self.rec(expr.index),
            {expr.aggregate.name: access_range}])

# }}}


# {{{ count

def add_assumptions_guard(kernel, pwqpolynomial):
    return GuardedPwQPolynomial(
            pwqpolynomial,
            kernel.assumptions.align_params(pwqpolynomial.space))


def count(kernel, set, space=None):
    if isinstance(kernel, TranslationUnit):
        kernel_names = [i for i, clbl in kernel.callables_table.items()
                if isinstance(clbl, CallableKernel)]
        if len(kernel_names) > 1:
            raise LoopyError()
        return count(kernel[kernel_names[0]], set, space)

    try:
        if space is not None:
            set = set.align_params(space)

        return add_assumptions_guard(kernel, set.card())
    except AttributeError:
        pass

    total_count = isl.PwQPolynomial.zero(
            set.space
            .drop_dims(dim_type.set, 0, set.dim(dim_type.set))
            .add_dims(dim_type.set, 1))

    set = set.make_disjoint()

    from loopy.isl_helpers import get_simple_strides

    for bset in set.get_basic_sets():
        bset_count = None
        bset_rebuilt = bset.universe(bset.space)

        bset_strides = get_simple_strides(bset, key_by="index")

        for i in range(bset.dim(isl.dim_type.set)):
            dmax = bset.dim_max(i)
            dmin = bset.dim_min(i)

            stride = bset_strides.get((dim_type.set, i))
            if stride is None:
                stride = 1

            length_pwaff = dmax - dmin + stride
            if space is not None:
                length_pwaff = length_pwaff.align_params(space)

            length = isl.PwQPolynomial.from_pw_aff(length_pwaff)
            length = length.scale_down_val(stride)

            if bset_count is None:
                bset_count = length
            else:
                bset_count = bset_count * length

            # {{{ rebuild check domain

            zero = isl.Aff.zero_on_domain(
                        isl.LocalSpace.from_space(bset.space))
            iname = isl.PwAff.from_aff(
                    zero.set_coefficient_val(isl.dim_type.in_, i, 1))
            dmin_matched = dmin.insert_dims(
                    dim_type.in_, 0, bset.dim(isl.dim_type.set))
            dmax_matched = dmax.insert_dims(
                    dim_type.in_, 0, bset.dim(isl.dim_type.set))
            for idx in range(bset.dim(isl.dim_type.set)):
                if bset.has_dim_id(isl.dim_type.set, idx):
                    dim_id = bset.get_dim_id(isl.dim_type.set, idx)
                    dmin_matched = dmin_matched.set_dim_id(
                            isl.dim_type.in_, idx, dim_id)
                    dmax_matched = dmax_matched.set_dim_id(
                            isl.dim_type.in_, idx, dim_id)

            bset_rebuilt = (
                    bset_rebuilt
                    & iname.le_set(dmax_matched)
                    & iname.ge_set(dmin_matched)
                    & (iname-dmin_matched).mod_val(stride).eq_set(zero))

            # }}}

        if bset_count is not None:
            total_count += bset_count

        is_subset = bset <= bset_rebuilt
        is_superset = bset >= bset_rebuilt

        if not (is_subset and is_superset):
            if is_subset:
                warn_with_kernel(kernel, "count_overestimate",
                        "Barvinok wrappers are not installed. "
                        "Counting routines have overestimated the "
                        "number of integer points in your loop "
                        "domain.")
            elif is_superset:
                warn_with_kernel(kernel, "count_underestimate",
                        "Barvinok wrappers are not installed. "
                        "Counting routines have underestimated the "
                        "number of integer points in your loop "
                        "domain.")
            else:
                warn_with_kernel(kernel, "count_misestimate",
                        "Barvinok wrappers are not installed. "
                        "Counting routines have misestimated the "
                        "number of integer points in your loop "
                        "domain.")

    return add_assumptions_guard(kernel, total_count)


def get_unused_hw_axes_factor(knl, callables_table, insn, disregard_local_axes):
    # FIXME: Multi-kernel support
    gsize, lsize = knl.get_grid_size_upper_bounds(callables_table)

    g_used = set()
    l_used = set()

    from loopy.kernel.data import GroupInameTag, LocalInameTag
    for iname in insn.within_inames:
        tags = knl.iname_tags_of_type(iname,
                              (LocalInameTag, GroupInameTag), max_num=1)
        if tags:
            tag, = tags
            if isinstance(tag, LocalInameTag):
                l_used.add(tag.axis)
            elif isinstance(tag, GroupInameTag):
                g_used.add(tag.axis)

    def mult_grid_factor(used_axes, sizes):
        result = get_kernel_zero_pwqpolynomial(knl) + 1

        for iaxis, size in enumerate(sizes):
            if iaxis not in used_axes:
                if isinstance(size, int):
                    result = result * size
                else:
                    result = result * isl.PwQPolynomial.from_pw_aff(
                            size.align_params(result.space))

        return result

    if disregard_local_axes:
        result = mult_grid_factor(g_used, gsize)
    else:
        result = mult_grid_factor(g_used, gsize) * mult_grid_factor(l_used, lsize)

    return add_assumptions_guard(knl, result)


def count_inames_domain(knl, inames):
    space = get_kernel_parameter_space(knl)
    if not inames:
        return add_assumptions_guard(knl,
                get_kernel_zero_pwqpolynomial(knl) + 1)

    inames_domain = knl.get_inames_domain(inames)
    domain = inames_domain.project_out_except(inames, [dim_type.set])
    return count(knl, domain, space=space)


def count_insn_runs(knl, callables_table, insn, count_redundant_work,
        disregard_local_axes=False):

    insn_inames = insn.within_inames

    if disregard_local_axes:
        from loopy.kernel.data import LocalInameTag
        insn_inames = frozenset(
                [iname for iname in insn_inames
                    if not knl.iname_tags_of_type(iname, LocalInameTag)])

    c = count_inames_domain(knl, insn_inames)

    if count_redundant_work:
        unused_fac = get_unused_hw_axes_factor(knl, callables_table,
                insn, disregard_local_axes=disregard_local_axes)
        return c * unused_fac
    else:
        return c


def _get_insn_count(knl, callables_table, insn_id, subgroup_size,
        count_redundant_work, count_granularity=CountGranularity.WORKITEM):
    insn = knl.id_to_insn[insn_id]

    if count_granularity is None:
        warn_with_kernel(knl, "get_insn_count_assumes_granularity",
                         "get_insn_count: No count granularity passed, "
                         "assuming %s granularity."
                         % (CountGranularity.WORKITEM))
        count_granularity = CountGranularity.WORKITEM

    if count_granularity == CountGranularity.WORKITEM:
        return count_insn_runs(
            knl, callables_table, insn,
            count_redundant_work=count_redundant_work,
            disregard_local_axes=False)

    ct_disregard_local = count_insn_runs(
            knl, callables_table, insn, disregard_local_axes=True,
            count_redundant_work=count_redundant_work)

    if count_granularity == CountGranularity.WORKGROUP:
        return ct_disregard_local
    elif count_granularity == CountGranularity.SUBGROUP:
        # {{{ compute workgroup_size

        from loopy.symbolic import aff_to_expr
        _, local_size = knl.get_grid_size_upper_bounds(callables_table)
        workgroup_size = 1
        if local_size:
            for size in local_size:
                if size.n_piece() != 1:
                    raise LoopyError("Workgroup size found to be genuinely "
                        "piecewise defined, which is not allowed in stats gathering")

                (valid_set, aff), = size.get_pieces()

                assert ((valid_set.n_basic_set() == 1)
                        and (valid_set.get_basic_sets()[0].is_universe()))

                s = aff_to_expr(aff)
                if not isinstance(s, int):
                    raise LoopyError("Cannot count insn with %s granularity, "
                                     "work-group size is not integer: %s"
                                     % (CountGranularity.SUBGROUP, local_size))
                workgroup_size *= s

        # }}}

        warn_with_kernel(knl, "insn_count_subgroups_upper_bound",
                "get_insn_count: when counting instruction %s with "
                "count_granularity=%s, using upper bound for work-group size "
                "(%d work-items) to compute sub-groups per work-group. When "
                "multiple device programs present, actual sub-group count may be "
                "lower." % (insn_id, CountGranularity.SUBGROUP, workgroup_size))

        from pytools import div_ceil
        return ct_disregard_local*div_ceil(workgroup_size, subgroup_size)

    else:
        # this should not happen since this is enforced in Op/MemAccess
        raise ValueError("get_insn_count: count_granularity '%s' is"
                "not allowed. count_granularity options: %s"
                % (count_granularity, [*CountGranularity.ALL, None]))

# }}}


# {{{ get_op_map

def _get_op_map_for_single_kernel(knl, callables_table,
        count_redundant_work,
        count_within_subscripts, subgroup_size, within):

    subgroup_size = _process_subgroup_size(knl, subgroup_size)

    kernel_rec = partial(_get_op_map_for_single_kernel,
            callables_table=callables_table,
            count_redundant_work=count_redundant_work,
            count_within_subscripts=count_within_subscripts,
            subgroup_size=subgroup_size, within=within)

    op_counter = ExpressionOpCounter(knl, callables_table, kernel_rec,
            count_within_subscripts)
    op_map = op_counter.new_zero_poly_map()

    from loopy.kernel.instruction import (
        Assignment,
        BarrierInstruction,
        CallInstruction,
        CInstruction,
        NoOpInstruction,
    )

    for insn in knl.instructions:
        if within(knl, insn):
            if isinstance(insn, (CallInstruction, Assignment)):
                ops = op_counter(insn.assignees) + op_counter(insn.expression)
                for key, val in ops.count_map.items():
                    count = _get_insn_count(knl, callables_table, insn.id,
                                subgroup_size, count_redundant_work,
                                key.count_granularity)
                    op_map = op_map + ToCountMap({key: val}) * count

            elif isinstance(
                    insn, (CInstruction, NoOpInstruction, BarrierInstruction)):
                pass
            else:
                raise NotImplementedError("unexpected instruction item type: '%s'"
                        % type(insn).__name__)

    return op_map


def get_op_map(program, count_redundant_work=False,
               count_within_subscripts=True, subgroup_size=None,
               entrypoint=None, within=None):

    """Count the number of operations in a loopy kernel.

    :arg knl: A :class:`loopy.LoopKernel` whose operations are to be counted.

    :arg count_redundant_work: Based on usage of hardware axes or other
        specifics, a kernel may perform work redundantly. This :class:`bool`
        flag indicates whether this work should be included in the count.
        (Likely desirable for performance modeling, but undesirable for code
        optimization.)

    :arg count_within_subscripts: A :class:`bool` specifying whether to
        count operations inside array indices.

    :arg subgroup_size: (currently unused) An :class:`int`, :class:`str`
        ``"guess"``, or *None* that specifies the sub-group size. An OpenCL
        sub-group is an implementation-dependent grouping of work-items within
        a work-group, analogous to an NVIDIA CUDA warp. subgroup_size is used,
        e.g., when counting a :class:`MemAccess` whose count_granularity
        specifies that it should only be counted once per sub-group. If set to
        *None* an attempt to find the sub-group size using the device will be
        made, if this fails an error will be raised. If a :class:`str`
        ``"guess"`` is passed as the subgroup_size, get_mem_access_map will
        attempt to find the sub-group size using the device and, if
        unsuccessful, will make a wild guess.

    :arg within: If not None, limit the result to matching contexts.
        See :func:`loopy.match.parse_match` for syntax.

    :return: A :class:`ToCountMap` of **{** :class:`Op` **:**
        :class:`islpy.PwQPolynomial` **}**.

        - The :class:`Op` specifies the characteristics of the arithmetic
          operation.

        - The :class:`islpy.PwQPolynomial` holds the number of operations of
          the kind specified in the key (in terms of the
          :class:`loopy.LoopKernel` parameter *inames*).

    Example usage::

        # (first create loopy kernel and specify array data types)

        op_map = get_op_map(knl)
        params = {"n": 512, "m": 256, "l": 128}
        f32add = op_map[Op(np.float32,
                           "add",
                           count_granularity=CountGranularity.WORKITEM)
                       ].eval_with_dict(params)
        f32mul = op_map[Op(np.float32,
                           "mul",
                           count_granularity=CountGranularity.WORKITEM)
                       ].eval_with_dict(params)

        # (now use these counts to, e.g., predict performance)

    """

    if entrypoint is None:
        if len(program.entrypoints) > 1:
            raise LoopyError("Must provide entrypoint")

        entrypoint = next(iter(program.entrypoints))

    assert entrypoint in program.entrypoints

    from loopy.preprocess import infer_unknown_types, preprocess_program
    program = preprocess_program(program)

    from loopy.match import parse_match
    within = parse_match(within)

    # Ordering restriction: preprocess might insert arguments to
    # make strides valid. Those also need to go through type inference.
    program = infer_unknown_types(program, expect_completion=True)

    return _get_op_map_for_single_kernel(
            program[entrypoint], program.callables_table,
            count_redundant_work=count_redundant_work,
            count_within_subscripts=count_within_subscripts,
            subgroup_size=subgroup_size,
            within=within)

# }}}


# {{{ subgoup size finding

def _find_subgroup_size_for_knl(knl):
    from loopy.target.pyopencl import PyOpenCLTarget
    if isinstance(knl.target, PyOpenCLTarget) and knl.target.device is not None:
        from pyopencl.characterize import get_simd_group_size
        subgroup_size_guess = get_simd_group_size(knl.target.device, None)
        warn_with_kernel(knl, "getting_subgroup_size_from_device",
                         "Device: %s. Using sub-group size given by "
                         "pyopencl.characterize.get_simd_group_size(): %s"
                         % (knl.target.device, subgroup_size_guess))
        return subgroup_size_guess
    else:
        return None


@memoize_method
def _process_subgroup_size(knl, subgroup_size_requested):

    if isinstance(subgroup_size_requested, int):
        return subgroup_size_requested
    else:
        # try to find subgroup_size
        subgroup_size_guess = _find_subgroup_size_for_knl(knl)

        if subgroup_size_requested is None:
            if subgroup_size_guess is None:
                # "guess" was not passed and either no target device found
                # or get_simd_group_size returned None
                raise ValueError("No sub-group size passed, no target device found. "
                                 "Either (1) pass integer value for subgroup_size, "
                                 "(2) ensure that kernel.target is PyOpenClTarget "
                                 "and kernel.target.device is set, or (3) pass "
                                 "subgroup_size='guess' and hope for the best.")
            else:
                return subgroup_size_guess

        elif subgroup_size_requested == "guess":
            if subgroup_size_guess is None:
                # unable to get subgroup_size from device, so guess
                subgroup_size_guess = 32
                warn_with_kernel(knl, "get_x_map_guessing_subgroup_size",
                                 "'guess' sub-group size passed, no target device "
                                 "found, wildly guessing that sub-group size is %d."
                                 % (subgroup_size_guess))
                return subgroup_size_guess
            else:
                return subgroup_size_guess
        else:
            raise ValueError("Invalid value for subgroup_size: %s. subgroup_size "
                             "must be integer, 'guess', or, if you're feeling "
                             "lucky, None." % (subgroup_size_requested))

# }}}


# {{{ get_mem_access_map

def _get_mem_access_map_for_single_kernel(knl, callables_table,
        count_redundant_work, subgroup_size, within):

    subgroup_size = _process_subgroup_size(knl, subgroup_size)

    kernel_rec = partial(_get_mem_access_map_for_single_kernel,
            callables_table=callables_table,
            count_redundant_work=count_redundant_work,
            subgroup_size=subgroup_size)

    access_counter_g = GlobalMemAccessCounter(
            knl, callables_table, kernel_rec)
    access_counter_l = LocalMemAccessCounter(
            knl, callables_table, kernel_rec)
    access_map = access_counter_g.new_zero_poly_map()

    from loopy.kernel.instruction import (
        Assignment,
        BarrierInstruction,
        CallInstruction,
        CInstruction,
        NoOpInstruction,
    )

    for insn in knl.instructions:
        if within(knl, insn):
            if isinstance(insn, (CallInstruction, Assignment)):
                insn_access_map = (
                            access_counter_g(insn.expression)
                            + access_counter_l(insn.expression)
                            ).with_set_attributes(direction="load")
                for assignee in insn.assignees:
                    insn_access_map = insn_access_map + (
                            access_counter_g(assignee)
                            + access_counter_l(assignee)
                            ).with_set_attributes(direction="store")

                for key, val in insn_access_map.count_map.items():
                    count = _get_insn_count(knl, callables_table, insn.id,
                                subgroup_size, count_redundant_work,
                                key.count_granularity)
                    access_map = access_map + ToCountMap({key: val}) * count

            elif isinstance(
                    insn, (CInstruction, NoOpInstruction, BarrierInstruction)):
                pass

            else:
                raise NotImplementedError("unexpected instruction item type: '%s'"
                        % type(insn).__name__)

    return access_map


def get_mem_access_map(program, count_redundant_work=False,
                       subgroup_size=None, entrypoint=None,
                       within=None):
    """Count the number of memory accesses in a loopy kernel.

    :arg knl: A :class:`loopy.LoopKernel` whose memory accesses are to be
        counted.

    :arg count_redundant_work: Based on usage of hardware axes or other
        specifics, a kernel may perform work redundantly. This :class:`bool`
        flag indicates whether this work should be included in the count.
        (Likely desirable for performance modeling, but undesirable for
        code optimization.)

    :arg subgroup_size: An :class:`int`, :class:`str` ``"guess"``, or
        *None* that specifies the sub-group size. An OpenCL sub-group is an
        implementation-dependent grouping of work-items within a work-group,
        analogous to an NVIDIA CUDA warp. subgroup_size is used, e.g., when
        counting a :class:`MemAccess` whose count_granularity specifies that it
        should only be counted once per sub-group. If set to *None* an attempt
        to find the sub-group size using the device will be made, if this fails
        an error will be raised. If a :class:`str` ``"guess"`` is passed as
        the subgroup_size, get_mem_access_map will attempt to find the
        sub-group size using the device and, if unsuccessful, will make a wild
        guess.

    :arg within: If not None, limit the result to matching contexts.
        See :func:`loopy.match.parse_match` for syntax.

    :return: A :class:`ToCountMap` of **{** :class:`MemAccess` **:**
        :class:`islpy.PwQPolynomial` **}**.

        - The :class:`MemAccess` specifies the characteristics of the memory
          access.

        - The :class:`islpy.PwQPolynomial` holds the number of memory accesses
          with the characteristics specified in the key (in terms of the
          :class:`loopy.LoopKernel` *inames*).

    Example usage::

        # (first create loopy kernel and specify array data types)

        params = {"n": 512, "m": 256, "l": 128}
        mem_map = get_mem_access_map(knl)

        f32_s1_g_ld_a = mem_map[MemAccess(
                                    mtype="global",
                                    dtype=np.float32,
                                    lid_strides={0: 1},
                                    gid_strides={0: 256},
                                    direction="load",
                                    variable="a",
                                    count_granularity=CountGranularity.WORKITEM)
                               ].eval_with_dict(params)
        f32_s1_g_st_a = mem_map[MemAccess(
                                    mtype="global",
                                    dtype=np.float32,
                                    lid_strides={0: 1},
                                    gid_strides={0: 256},
                                    direction="store",
                                    variable="a",
                                    count_granularity=CountGranularity.WORKITEM)
                               ].eval_with_dict(params)
        f32_s1_l_ld_x = mem_map[MemAccess(
                                    mtype="local",
                                    dtype=np.float32,
                                    lid_strides={0: 1},
                                    gid_strides={0: 256},
                                    direction="load",
                                    variable="x",
                                    count_granularity=CountGranularity.WORKITEM)
                               ].eval_with_dict(params)
        f32_s1_l_st_x = mem_map[MemAccess(
                                    mtype="local",
                                    dtype=np.float32,
                                    lid_strides={0: 1},
                                    gid_strides={0: 256},
                                    direction="store",
                                    variable="x",
                                    count_granularity=CountGranularity.WORKITEM)
                               ].eval_with_dict(params)

        # (now use these counts to, e.g., predict performance)

    """

    if entrypoint is None:
        if len(program.entrypoints) > 1:
            raise LoopyError("Must provide entrypoint")

        entrypoint = next(iter(program.entrypoints))

    assert entrypoint in program.entrypoints

    from loopy.preprocess import infer_unknown_types, preprocess_program

    program = preprocess_program(program)

    from loopy.match import parse_match
    within = parse_match(within)

    # Ordering restriction: preprocess might insert arguments to
    # make strides valid. Those also need to go through type inference.
    program = infer_unknown_types(program, expect_completion=True)

    return _get_mem_access_map_for_single_kernel(
            program[entrypoint], program.callables_table,
            count_redundant_work=count_redundant_work,
            subgroup_size=subgroup_size,
            within=within)

# }}}


# {{{ get_synchronization_map

def _get_synchronization_map_for_single_kernel(knl, callables_table,
        subgroup_size=None):

    knl = lp.get_one_linearized_kernel(knl, callables_table)

    from loopy.schedule import (
        Barrier,
        CallKernel,
        EnterLoop,
        LeaveLoop,
        ReturnFromKernel,
        RunInstruction,
    )

    kernel_rec = partial(_get_synchronization_map_for_single_kernel,
            callables_table=callables_table,
            subgroup_size=subgroup_size)

    sync_counter = CounterBase(knl, callables_table, kernel_rec)
    sync_map = sync_counter.new_zero_poly_map()

    iname_list = []

    for sched_item in knl.linearization:
        if isinstance(sched_item, EnterLoop):
            if sched_item.iname:  # (if not empty)
                iname_list.append(sched_item.iname)
        elif isinstance(sched_item, LeaveLoop):
            if sched_item.iname:  # (if not empty)
                iname_list.pop()

        elif isinstance(sched_item, Barrier):
            sync_map = sync_map + ToCountMap(
                    {Sync(
                        "barrier_%s" % sched_item.synchronization_kind,
                        knl.name): count_inames_domain(knl, frozenset(iname_list))})

        elif isinstance(sched_item, RunInstruction):
            pass

        elif isinstance(sched_item, CallKernel):
            sync_map = sync_map + ToCountMap(
                    {Sync("kernel_launch", knl.name):
                        count_inames_domain(knl, frozenset(iname_list))})

        elif isinstance(sched_item, ReturnFromKernel):
            pass

        else:
            raise LoopyError("unexpected schedule item: %s"
                    % type(sched_item).__name__)

    return sync_map


def get_synchronization_map(program, subgroup_size=None, entrypoint=None):
    """Count the number of synchronization events each work-item encounters in
    a loopy kernel.

    :arg knl: A :class:`loopy.LoopKernel` whose barriers are to be counted.

    :arg subgroup_size: (currently unused) An :class:`int`, :class:`str`
        ``"guess"``, or *None* that specifies the sub-group size. An OpenCL
        sub-group is an implementation-dependent grouping of work-items within
        a work-group, analogous to an NVIDIA CUDA warp. subgroup_size is used,
        e.g., when counting a :class:`MemAccess` whose count_granularity
        specifies that it should only be counted once per sub-group. If set to
        *None* an attempt to find the sub-group size using the device will be
        made, if this fails an error will be raised. If a :class:`str`
        ``"guess"`` is passed as the subgroup_size, get_mem_access_map will
        attempt to find the sub-group size using the device and, if
        unsuccessful, will make a wild guess.

    :return: A dictionary mapping each type of synchronization event to an
        :class:`islpy.PwQPolynomial` holding the number of events per
        work-item.

        Possible keys include ``barrier_local``, ``barrier_global``
        (if supported by the target) and ``kernel_launch``.

    Example usage::

        # (first create loopy kernel and specify array data types)

        sync_map = get_synchronization_map(knl)
        params = {"n": 512, "m": 256, "l": 128}
        barrier_ct = sync_map["barrier_local"].eval_with_dict(params)

        # (now use this count to, e.g., predict performance)

    """
    if entrypoint is None:
        if len(program.entrypoints) > 1:
            raise LoopyError("Must provide entrypoint")

        entrypoint = next(iter(program.entrypoints))

    assert entrypoint in program.entrypoints
    from loopy.preprocess import infer_unknown_types, preprocess_program

    program = preprocess_program(program)
    # Ordering restriction: preprocess might insert arguments to
    # make strides valid. Those also need to go through type inference.
    program = infer_unknown_types(program, expect_completion=True)

    return _get_synchronization_map_for_single_kernel(
            program[entrypoint], program.callables_table,
            subgroup_size=subgroup_size)

# }}}


# {{{ gather_access_footprints

def _gather_access_footprints_for_single_kernel(kernel, ignore_uncountable):
    write_footprints = []
    read_footprints = []

    for insn in kernel.instructions:
        if not isinstance(insn, MultiAssignmentBase):
            warn_with_kernel(kernel, "count_non_assignment",
                    "Non-assignment instruction encountered in "
                    "gather_access_footprints, not counted")
            continue

        insn_inames = kernel.insn_inames(insn)
        inames_domain = kernel.get_inames_domain(insn_inames)
        domain = (inames_domain.project_out_except(insn_inames,
                                                   [dim_type.set]))

        afg = AccessFootprintGatherer(kernel, domain,
                ignore_uncountable=ignore_uncountable)

        write_footprints.append(afg(insn.assignees))
        read_footprints.append(afg(insn.expression))

    return write_footprints, read_footprints


def gather_access_footprints(program, ignore_uncountable=False, entrypoint=None):
    """Return a dictionary mapping ``(var_name, direction)`` to
    :class:`islpy.Set` instances capturing which indices of each the array
    *var_name* are read/written (where *direction* is either ``read`` or
    ``write``.

    :arg ignore_uncountable: If *False*, an error will be raised for accesses
        on which the footprint cannot be determined (e.g. data-dependent or
        nonlinear indices)
    """

    if entrypoint is None:
        if len(program.entrypoints) > 1:
            raise LoopyError("Must provide entrypoint")

        entrypoint = next(iter(program.entrypoints))

    assert entrypoint in program.entrypoints

    # FIMXE: works only for one callable kernel till now.
    if len([in_knl_callable for in_knl_callable in
        program.callables_table.values() if isinstance(in_knl_callable,
            CallableKernel)]) != 1:
        raise NotImplementedError("Currently only supported for program with "
            "only one CallableKernel.")

    from loopy.preprocess import infer_unknown_types, preprocess_program

    program = preprocess_program(program)
    # Ordering restriction: preprocess might insert arguments to
    # make strides valid. Those also need to go through type inference.
    program = infer_unknown_types(program, expect_completion=True)

    write_footprints = []
    read_footprints = []

    write_footprints, read_footprints = _gather_access_footprints_for_single_kernel(
            program[entrypoint], ignore_uncountable)

    write_footprints = AccessFootprintGatherer.combine(write_footprints)
    read_footprints = AccessFootprintGatherer.combine(read_footprints)

    result = {}

    for vname, footprint in write_footprints.items():
        result[vname, "write"] = footprint

    for vname, footprint in read_footprints.items():
        result[vname, "read"] = footprint

    return result


def gather_access_footprint_bytes(program, ignore_uncountable=False):
    """Return a dictionary mapping ``(var_name, direction)`` to
    :class:`islpy.PwQPolynomial` instances capturing the number of bytes  are
    read/written (where *direction* is either ``read`` or ``write`` on array
    *var_name*

    :arg ignore_uncountable: If *True*, an error will be raised for accesses on
        which the footprint cannot be determined (e.g. data-dependent or
        nonlinear indices)
    """

    from loopy.preprocess import infer_unknown_types, preprocess_program
    kernel = infer_unknown_types(program, expect_completion=True)

    from loopy.kernel import KernelState
    if kernel.state < KernelState.PREPROCESSED:
        kernel = preprocess_program(program)

    result = {}
    fp = gather_access_footprints(kernel,
                                  ignore_uncountable=ignore_uncountable)

    for key, var_fp in fp.items():
        vname, _direction = key

        var_descr = kernel.get_var_descriptor(vname)
        bytes_transferred = (
                int(var_descr.dtype.numpy_dtype.itemsize)
                * count(kernel, var_fp))
        if key in result:
            result[key] += bytes_transferred
        else:
            result[key] = bytes_transferred

    return result

# }}}

# vim: foldmethod=marker



__copyright__ = "Copyright (C) 2021 University of Illinois Board of Trustees"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import numpy as np

import pyopencl as cl
from pyopencl.tools import (  # noqa: F401
    pytest_generate_tests_for_pyopencl as pytest_generate_tests,
)

import loopy as lp


def test_two_kernel_fusion(ctx_factory):
    """
    A simple fusion test of two sets of instructions.
    """

    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)

    knla = lp.make_kernel(
        "{[i]: 0<=i<10}",
        """
            out[i] = i
        """
    )
    knlb = lp.make_kernel(
        "{[j]: 0<=j<10}",
        """
            out[j] = j+100
        """
    )
    knl = lp.fuse_kernels([knla, knlb], data_flow=[("out", 0, 1)])
    _evt, (out,) = knl(queue)
    np.testing.assert_allclose(out.get(), np.arange(100, 110))


def test_write_block_matrix_fusion(ctx_factory):
    """
    A slightly more complicated fusion test, where all
    sub-kernels write into the same global matrix, but
    in well-defined separate blocks. This tests makes sure
    data flow specification is preserved during fusion for
    matrix-assembly-like programs.
    """

    ctx = ctx_factory()
    queue = cl.CommandQueue(ctx)
    rng = np.random.default_rng(seed=42)

    def init_global_mat_prg():
        return lp.make_kernel(
            [
                "{[idof]: 0 <= idof < n}",
                "{[jdof]: 0 <= jdof < m}"
            ],
            """
                result[idof, jdof]  = 0 {id=init}
            """,
            [
                lp.GlobalArg("result", None,
                    shape="n, m",
                    offset=lp.auto),
                lp.ValueArg("n, m", np.int32),
                "...",
            ],
            options=lp.Options(
                return_dict=True
            ),
            default_offset=lp.auto,
            name="init_a_global_matrix",
        )

    def write_into_mat_prg():
        return lp.make_kernel(
            [
                "{[idof]: 0 <= idof < ndofs}",
                "{[jdof]: 0 <= jdof < mdofs}"
            ],
            """
                result[offset_i + idof, offset_j + jdof] = mat[idof, jdof]
            """,
            [
                lp.GlobalArg("result", None,
                    shape="n, m",
                    offset=lp.auto),
                lp.ValueArg("n, m", np.int32),
                lp.GlobalArg("mat", None,
                    shape="ndofs, mdofs",
                    offset=lp.auto),
                lp.ValueArg("offset_i", np.int32),
                lp.ValueArg("offset_j", np.int32),
                "...",
            ],
            options=lp.Options(
                return_dict=True
            ),
            default_offset=lp.auto,
            name="write_into_global_matrix",
        )

    # Construct a 2x2 diagonal matrix with
    # random 5x5 blocks on the block-diagonal,
    # and zeros elsewhere
    n = 10
    block_n = 5
    mat1 = rng.normal(size=(block_n, block_n))
    mat2 = rng.normal(size=(block_n, block_n))
    answer = np.block([[mat1, np.zeros((block_n, block_n))],
                      [np.zeros((block_n, block_n)), mat2]])
    kwargs = {"n": n, "m": n}

    # Do some renaming of individual programs before fusion
    kernels = [init_global_mat_prg()]
    for idx, (offset, mat) in enumerate([(0, mat1), (block_n, mat2)]):
        knl = lp.rename_argument(write_into_mat_prg(), "mat", f"mat_{idx}")
        kwargs[f"mat_{idx}"] = mat

        for iname in knl.default_entrypoint.all_inames():
            knl = lp.rename_iname(knl, iname, f"{iname}_{idx}")

        knl = lp.rename_argument(knl, "ndofs", f"ndofs_{idx}")
        knl = lp.rename_argument(knl, "mdofs", f"mdofs_{idx}")
        kwargs[f"ndofs_{idx}"] = block_n
        kwargs[f"mdofs_{idx}"] = block_n

        knl = lp.rename_argument(knl, "offset_i", f"offset_i_{idx}")
        knl = lp.rename_argument(knl, "offset_j", f"offset_j_{idx}")
        kwargs[f"offset_i_{idx}"] = offset
        kwargs[f"offset_j_{idx}"] = offset

        kernels.append(knl)

    fused_knl = lp.fuse_kernels(
        kernels,
        data_flow=[("result", 0, 1), ("result", 1, 2)],
    )
    fused_knl = lp.add_nosync(
        fused_knl,
        "global",
        "writes:result",
        "writes:result",
        bidirectional=True,
        force=True
    )
    _evt, result = fused_knl(queue, **kwargs)
    result = result["result"]
    np.testing.assert_allclose(result, answer)


if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from pytest import main
        main([__file__])
